/******************************************************************************
 * $Id: module.c,v 1.22 2001/09/03 04:47:35 jimb Exp $
 * Author		: Jim Brooks	<jim@jimbrooks.org>
 * Description	: Functions for managing source code modules (ie files).
 ******************************************************************************/

#include "header.h"

/******************************************************************************
 * Return a pointer to a buffer that holds a source code module (ie file).
 * Adds (or updates) a HASH_ITEM_MODULE struct to the module hash table.
 *
 * The buffer has a null char appended that is NOT accounted
 * by the returned module size.
 *
 * If the module has not been loaded, then this will load it.
 *
 * To free the buffer, call FreeModuleBuffer(), not g_free()!!!
 *
 * Returns : Pointer to memory buffer allocated by ** g_malloc() **
 *           -or- NULL upon failure.
 *
 * Parms   : pModuleName
 *           Name of module.
 *
 *           pModuleSize (OUT)
 *           Size of module (undefined if module could not be loaded).
 *           This does NOT account a null char that was appended to buffer.
 *
 *           pHashItemModule (GLOBAL/OUT)
 *           Updated.
 ******************************************************************************/
gpointer
ModuleBuffer( gchar* pModuleName,
              long*  pModuleSize /* OUT */ )
{
  /*
   * If the sought module has been loaded into a buffer,
   * then it will have a hash item that has a non-NULL pointer to the buffer.
   */
   pHashItemModule = (P_HASH_ITEM_MODULE)g_hash_table_lookup( pHashModules,
                                                        (gconstpointer)pModuleName );
   if ( pHashItemModule  &&  pHashItemModule->pModuleBuffer )
   {
     /*
      * Output the pointer to the buffer, and the size of the buffer.
      */
#if 0
     /* A buffer can legitimately be empty */
      g_assert( pHashItemModule->moduleSize );
#endif
      *pModuleSize = pHashItemModule->moduleSize;
      return pHashItemModule->pModuleBuffer;
   }

  /*
   * If the module has never been loaded, then create a hash item struct.
   * Or, its hash item is still allocated but the module was freed.
   */
   if ( pHashItemModule == NULL )
   {
      pHashItemModule = (P_HASH_ITEM_MODULE) g_malloc0( sizeof(HASH_ITEM_MODULE) );
      g_return_val_if_fail( pHashItemModule, NULL );
#ifdef HYPERSRC_PEDANTIC
      g_return_val_if_fail( sizeof(HASH_ITEM_MODULE), NULL );
#endif
   }

  /*
   * Load the file into a buffer.
   * LoadFile() will append a null to buffer that is not
   * accounted in the returned size.
   * Consequently, fill the hash item struct.
   */
   pHashItemModule->pModuleBuffer = LoadFile( pModuleName, pModuleSize /*OUT*/ );
   if ( pHashItemModule->pModuleBuffer == NULL )
   {
     /*
      * Oops, couldn't open the module.
      */
      PrintStatusbarThirds( "Could not open module '", pModuleName, "'." );
      if ( FileSize( pModuleName ) == 0 )
         Warning( "module %s is empty. ", pModuleName );
      else
         Warning( "couldn't open module %s ", pModuleName );
      return NULL;
   }
   else
   {
     /*
      * Does this appear to be a MSDOS file?
      */
      if ( memchr( pHashItemModule->pModuleBuffer, CHAR_CR, min(100,*pModuleSize) ) )
      {
         char* pBufferNew;
         int   bufferNewSize;

        /*
         * Create a new buffer minus CR chars, then free old buffer.
         */
         if ( FilterMsdosCr( pHashItemModule->pModuleBuffer,
                             *pModuleSize,
                             &pBufferNew,
                             &bufferNewSize ) == FUNC_OK )
         {
            g_free( pHashItemModule->pModuleBuffer );
            pHashItemModule->pModuleBuffer = pBufferNew;
            pHashItemModule->moduleSize = *pModuleSize = bufferNewSize;
         }
         /* Oops, FilterMsdosCr() failed.  Continue without filtration of CR chars. */
      }
   }
   pHashItemModule->moduleSize = *pModuleSize;

#ifdef HYPERSRC_PEDANTIC
  /*
   * Check if a null char exists within the buffer,
   * because code which parses buffers would malfunction.
   * LoadFile() has appended the buffer with a null char.
   */
 {
   char* pNull = strchr( pHashItemModule->pModuleBuffer, '\0' );
   if ( pNull < (char*)pHashItemModule->pModuleBuffer + *pModuleSize )
   {
      Warning( "A null char found in middle of file %s \n", pModuleName );
      return NULL;
   }
 }
#endif

  /*
   * Completely fill the hash item.
   */
#if 0
   /* --- keep unchanged --- */
   pHashItemModule->pGslistFuncDef = NULL;
#endif
   pHashItemModule->pModuleName = CloneString( pModuleName );

  /*
   * Insert/replace the hash item in the hash table.
   */
   g_hash_table_insert( pHashModules,
                        (gpointer)pModuleName,
                        (gpointer)pHashItemModule );

  /*
   * (The hash item struct will remain allocated.)
   */

  /*
   * Return a pointer to the buffer that holds the module.
   * (The size of the buffer was output to "pModuleSize".)
   */
   return pHashItemModule->pModuleBuffer;
}

/******************************************************************************
 * Free buffer that holds a module.
 * Updates the corresponding module hash item (it is left in the hash table).
 *
 * Parms   : pModuleName
 *           Name of module.
 *
 *         : pHashItemModule->pModuleBuffer (OUT/GLOBAL)
 *           pHashItemModule->moduleSize    (OUT/GLOBAL)
 *           Updated to indicate module was freed.
 ******************************************************************************/
void
FreeModuleBuffer( gchar* pModuleName )
{
   P_HASH_ITEM_MODULE	pHashItem;

  /*
   * Skip if name indicates if no module in text widget.
   */
   if (    pModuleName == NULL
        || pModuleName[0] == '\0' )
      return;

  /*
   * If the sought module has been loaded into a buffer,
   * then it will have a hash item that has a non-NULL pointer to the buffer.
   */
   pHashItem = (P_HASH_ITEM_MODULE)g_hash_table_lookup( pHashModules,
                                                        (gconstpointer)pModuleName );
   if ( pHashItem  &&  pHashItem->pModuleBuffer )
   {
     /*
      * Free the buffer.
      * Record in hash item that buffer was freed.
      */
      DeleteString( pHashItem->pModuleName );
      pHashItem->pModuleName = NULL;

      g_free( pHashItem->pModuleBuffer );
      pHashItem->pModuleBuffer = NULL;

      pHashItem->moduleSize    = 0;

     /*
      * --- leave pGslistFuncDef unchanged ---
      */
   }

   return;
}

/******************************************************************************
 * Returns a hash value from a key.
 ******************************************************************************/
guint
HashModuleKey( gconstpointer pKey )
{
#if 1
   return g_str_hash( pKey );
#else
   static hashModuleKey = 0;

   return ++hashModuleKey;
#endif
}

/******************************************************************************
 * The comparsion function for the module hash.
 ******************************************************************************/
gint
HashModuleCmp( gconstpointer pName1,
               gconstpointer pName2 )
{
#if 1
   return g_str_equal( pName1,
                       pName2 );
#else
   return ! CompareStrings( (char*) pName1,
                            (char*) pName2 );
#endif
}

/*****************************************************************************
 * Create a sorted glist of module names from the module hash table.
 *
 * Parms:		pHashModules (GLOBAL)
 *
 *				pGlistModuleNames (OUT/GLOBAL)
 *****************************************************************************/
void
CreateModuleNameGlist( void )
{
   g_assert( pHashModules ); /* must be created first */

  /*
   * Erase previous module names glist.
   */
   if ( pGlistModuleNames )
   {
      g_list_free( pGlistModuleNames );

     /*
      * Reset to NULL to cause a new glist to be created.
      */
      pGlistModuleNames = NULL;
   }

  /*
   * The modules in the hash are unsorted, so create a glist of sorted module names.
   */
   g_hash_table_foreach( pHashModules,
                         AddSortedModuleGlist,
                         (gpointer)&pGlistModuleNames /* out */ );

  return;
}

/******************************************************************************
 * From an item in the module hash, add a module name to a sorted glist.
 *
 * Parms   : pModuleName (hash key)
 *           Pointer to the module name.
 *
 *         : pUnused (hash value)
 *
 *         : ppGlist (IN/OUT)
 *           Pointer^2 to the glist containing module names.
 ******************************************************************************/
void
AddSortedModuleGlist( gpointer pModuleName,
                      gpointer pUnused,
                      gpointer ppGlist )
{
   *(GList**)ppGlist = g_list_insert_sorted( *(GList**)ppGlist,
                                             pModuleName,
                                             HashStrcmp );
   return;
}
