/*****************************************************************************
 * $Id: parser.l,v 1.31 2001/09/03 16:25:08 jimb Exp $
 * Author		: Jim Brooks	<jim@jimbrooks.org>
 * Description	: Lexes function definitions and function calls in C/C++ code.
 *****************************************************************************/

%{

#define PARSER_L

#if 0
#define DEBUG
#endif

#undef yywrap
#define YY_NEVER_INTERACTIVE	1
#define YY_NO_UNPUT			/* shutup warning: `yyunput' defined but not used */

/*
 * Bits for discarding tokens.
 */
#define DISCARD_C_COMMENT				1
#define DISCARD_CPP_COMMENT				2
#define DISCARD_QUOTED					4
#define DISCARD_DEFINE					8
#define DISCARD_C_COMMENT_OR_QUOTED		(DISCARD_C_COMMENT | DISCARD_QUOTED)

#include "y.tab.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "parser.h"
#include "str_heap.h"

extern STR_HEAP strHeap;	/* created at parser.y */
extern int yylex();
extern int parserMode;
extern int parsingFuncDef;
extern int parsedLeftParOfFuncDef;
extern int parsingCallParms;
extern int betweenParmsAndBody;
#if 1
extern char parserInputName[];
#else
extern char* parserInputName; /* --- this caused a segfault, why? --- */
#endif

/*
 * Vars.
 */
static int tokenPassed;

/*
 * Vars which must be reset by LexReset().
 */
       int lineLex				= 1;
static int discard				= 0;		/* this is a bit-field */
static int parentheses			= 0;
static int braces				= 0;
static int prevToken			= 0;		/* possibly changed by LexToken() as another token */
static int prevTokenPassed		= 0;		/* early value of token as it was when passed to LexToken() */
static int cast					= 0;
static int mismatchBraces		= 0;
static int mismatchParentheses	= 0;
static int ignoreParenthesis	= 0;

/*
 * For non-lexical internal errors caused bugs, insufficent memory, bad pointers, etc.
 */
#ifdef LEX_YACC_QUIET
#define LEX_INTERNAL_ERROR_MSG( str ) { ; }
#else
#define LEX_INTERNAL_ERROR_MSG( str ) \
{ \
   printf( "lex INTERNAL ERROR: %s \n", str ); \
}
#endif

/*
 * For lexical errors present in the input being lexed.
 */
void LEXICAL_ERROR_MSG( char* str )
{
#ifdef LEX_YACC_QUIET
   return;
#endif

   if ( parserInputName[0] )
      printf( "%s:[%d]", parserInputName, lineLex );
   else
      printf( "[%d]", lineLex );

   printf( " lexical error" );
   if ( str[0] ) printf( ": %s", str ); 

   printf( "\n" );
}

/*
 * What to do if an internal error happens.
 */
#define LEX_ABORT() exit(1);

%}


/*===========================================================================*/

/*****************************************************************************
 * regex notes:  "." means any char except newline.
 *****************************************************************************/

%%

\'\"\'		|			/* Ignore a double-quote enclosed by single-quotes ('"' or '\"'). */
\'\\\"\'				{ return LexToken( yytext, TOKEN_TEXT ); }

"#endif"				{ /* Try to recover from an apparently mismatched parenthesis/brace */
                          /* that straddles '#'if/'#'else/'#'endif but would appear matched by cpp. */
                          LexErrorRecover();
                          return LexToken( yytext, TOKEN_TEXT );
                        }

\"[ ]*\#if[ ]*\"	|	/* Ignore keywords within quotes (strings) */
\"[ ]*\#elif[ ]*\"	|
\"[ ]*\#pragma[ ]*\"	{ return LexToken( yytext, TOKEN_TEXT ); }

\#if.*$					{ ; /* Ignore, eg "#if (ARCH==X86) && (SMP)" would confuse parser. */ }
\#elif.*$				{ ; }
\#pragma.*$				{ ; }

\#define				{ /* Discard tokens until last line of #define is encountered		*/
                          /* which is the next newline with a continuation char '\'.		*/
						  /* #define macros can contain parentheses which confuse parser.	*/
                          if ( !discard ) discard |= DISCARD_DEFINE; }
\\[ ]*\n				{ /* To discard tokens during a #define, don't let a newline	*/
                          /* with a continuation char '\' reach the below lex rule that	*/
                          /* clears DISCARD_DEFINE.										*/
                          lineLex++; }

if[ \t]*\(		|
for[ \t]*\(		|
while[ \t]*\(	|
switch[ \t]*\(			{ /* Filter-out a keyword's parentheses so that a function call within */
                          /* a keyword's parentheses [eg "while ( foo() )"] will be parsed correctly. */
                          ignoreParenthesis = TRUE;	}

sizeof\([^\)]*\)		{ /* Ignore "sizeof()". */
                          /* Don't greedily lex all right parentheses, eg: "if (x==sizeof(x))". */
                          return LexToken( yytext, TOKEN_TEXT ); }

\/\/					{ /* Ignore "//" if it's enclosed by a C comment or quotes. */
                          if ( (discard & DISCARD_C_COMMENT_OR_QUOTED) == 0 )
                             discard |= DISCARD_CPP_COMMENT;
						}

"/*"					{ discard |=  DISCARD_C_COMMENT; }

"*/"					{ discard &= ~DISCARD_C_COMMENT; }

\"						{ /* Ignore quotes enclosed in comments. */ 
                          if ( ! (discard & (DISCARD_C_COMMENT|DISCARD_CPP_COMMENT)) )
                          {
                             discard ^= DISCARD_QUOTED;
                             if ( ! (discard & DISCARD_QUOTED) ) return TOKEN_TEXT;
                          }
                        }

[a-zA-Z_]	|
[a-zA-Z_][a-zA-Z_0-9]+	{ return LexToken( yytext, TOKEN_ID ); }

\}[a-zA-Z0-9_ ]*;		{ /* Filter-out a right brace that is part of the end	*/
                          /* of a struct/union (including a typedef-ed struct).	*/
                          /* It would resemble:  "};"  or  "} type;"			*/
                          return LexToken( yytext, TOKEN_IGNORE ); }

						/* Only parse a brace on beginning of line and it belongs to a func def. */
^\{						{ return LexToken( yytext, parsedLeftParOfFuncDef ? yytext[0] : TOKEN_TEXT ); }
^\}						{ int res; /* End of a func def is a good point to recover lex from any error. */
                          res  = LexToken( yytext, parsedLeftParOfFuncDef ? yytext[0] : TOKEN_TEXT );
                          LexErrorRecover();
                          return res; }
"{"						{ return LexToken( yytext, TOKEN_TEXT ); }
"}"						{ return LexToken( yytext, TOKEN_TEXT ); }

[\(\)\,]				{ return LexToken( yytext, yytext[0] ); }

[ \t\r]+				{ ; }

\n						{ lineLex++;
                          if ( discard & DISCARD_CPP_COMMENT ) discard &= ~DISCARD_CPP_COMMENT;
                          if ( discard & DISCARD_DEFINE      ) discard &= ~DISCARD_DEFINE; }

.						{ return LexToken( yytext, TOKEN_TEXT ); }

%%


/*----------------------------------------------------------------------------
 * public functions
 *---------------------------------------------------------------------------*/


/*****************************************************************************
 * Tell lex to use a buffer.
 *
 * Returns:		Buffer's handle (really a pointer to YY_BUFFER_STATE).
 *****************************************************************************/
void*
LexBufferUse( char* buf )
{
   LexReset();
   yyrestart( NULL );
   return (void*) yy_scan_string( buf );
}

/*****************************************************************************
 * Discard buffer.
 *
 * Parms:		hnd
 *				Pass the handle returned by LexInputUse().
 *****************************************************************************/
void
LexBufferDiscard( void* hnd )
{
#ifdef HYPERSRC_PEDANTIC
   if ( hnd == NULL ) { LEX_INTERNAL_ERROR_MSG( "LexBufferDiscard() was passed NULL" ); return; }
#endif
   yy_delete_buffer( (YY_BUFFER_STATE) hnd );
}


/*----------------------------------------------------------------------------
 * internal functions
 *---------------------------------------------------------------------------*/


#define RETURN_TOKEN_ERROR() \
{ \
   if ( parserMode == PARSER_MODE_HYPERSRC ) \
      return TOKEN_ERROR; \
   else \
      LEX_ABORT(); \
}

#ifdef DEBUG
#define PRINT_TOKEN( lextoken ) \
{ \
   if ( ! discard ) \
   { \
      printf( "token = %d ", token ); \
      if ( token < 100 ) printf( " " ); \
      printf( " '%s' \n", pLexeme ); \
   } \
}
#else
#define PRINT_TOKEN( lextoken )
#endif /* DEBUG */

#define RETURN_TOKEN( lextoken ) \
{ \
   PRINT_TOKEN( lextoken ); \
   prevTokenPassed = tokenPassed; /* remember token before (if) LexToken() changed it */ \
   return prevToken = lextoken; \
} 

/*****************************************************************************
 * Process a lexeme/token.
 *
 * - record lexeme in a heap of strings
 * - assist parser by essentially preprocesing some tokens
 * - return a token to parser
 *****************************************************************************/
static int
LexToken( char* pLexeme, int token )
{
#ifdef HYPERSRC_PEDANTIC
   if ( strHeap.pBase == NULL  ||  strHeap.size == 0 )
   {
      LEX_INTERNAL_ERROR_MSG( "string heap hasn't been created yet." );
      LEX_ABORT();
   }
#endif

  /*
   * Record value of token before it might be changed.
   */
   tokenPassed = token;

  /*
   * Add this lexeme to a heap of strings.
   */
#ifndef HYPERSRC_PEDANTIC
   yylval.str = StrHeapAppendFast( &strHeap, pLexeme, strlen(pLexeme) );
#else
   yylval.str = StrHeapAppend(     &strHeap, pLexeme, strlen(pLexeme) );
   if ( yylval.str == NULL )
   {
      LEX_INTERNAL_ERROR_MSG( "StrHeapAppend() failed in parser.l" );
      LEX_ABORT();
   }
#endif

  /*
   * Discard tokens within comments or quotes.
   */
   if ( discard ) RETURN_TOKEN( TOKEN_TEXT );

  /*
   * Constructing a yacc rule to parse "int a, b, c;" is difficult.
   * This is an ugly alternative.
   */
   if ( token == ','  &&  parentheses == 0 )
      RETURN_TOKEN( TOKEN_TEXT );

  /*
   * Kludge to help the parser identify a function prototype.
   * A function prototype is distinguished from a function definition
   * by the sequential tokens ");" lexed outside of braces.
   */
   if ( prevTokenPassed == ')'  &&  pLexeme[0] == ';'  &&  pLexeme[1] == 0  &&  braces == 0 )
      RETURN_TOKEN( TOKEN_KLUDGE_FUNC_PROTO_END );

  /*
   * If the token is an TOKEN_ID but is part of a cast,
   * then change the token to TOKEN_TEXT to avoid confusing the parser.
   */
   if ( cast  &&  token == TOKEN_ID )
   {
      DebugMsg( "changing TOKEN_ID to TOKEN_TEXT" );
      RETURN_TOKEN( TOKEN_TEXT );
   }

  /*
   * To simplify the parser rules, ignore nested parentheses and type casts.
   */
   if ( token == '(' )
   {
     /*
      * So that the parser can handle K&R parameter types that contain parentheses,
      * ignore parentheses between ')' and '{' in beginning of a func def.
      */
      if ( betweenParmsAndBody )
         RETURN_TOKEN( TOKEN_TEXT );

      ++parentheses;
      if ( parentheses < 0 )
      {
         mismatchParentheses = TRUE;
         LEXICAL_ERROR_MSG( "parenthesis mismatch" );
         RETURN_TOKEN_ERROR();
      }

     /*
      * Function calls are assumed to begin with an identifier.
      * If this token is '(' but the previous token wasn't an identifier,
      * then assume it is the beginning of a type cast.
      * Ignore a cast because its syntactic similiarity to a function call
      * would confuse the function call yacc rules.
      */
      if ( prevToken != TOKEN_ID  &&  ! cast )
      {
         cast = parentheses;		/* remember the parenthesis level */
         RETURN_TOKEN( TOKEN_TEXT );
      }

      if ( parentheses == 1  ||  parsingCallParms )
      { RETURN_TOKEN( '(' ); }		/* probably belongs to a function def/call */
      else
      { RETURN_TOKEN( TOKEN_TEXT ); }		/* ignore this nested parenthesis */
   }
   else if ( token == ')' )
   {
     /*
      * So that the parser can handle K&R parameter types that contain parentheses,
      * ignore parentheses between ')' and '{' in beginning of a func def.
      */
      if ( betweenParmsAndBody )
         RETURN_TOKEN( TOKEN_TEXT );

      --parentheses;
      if ( parentheses < 0 )
      {
        /*
         * If the parentheses of a keyword was previously parsed --eg for ()--
         * then correct the parenthesis level.
         */
         if ( ignoreParenthesis  &&  parentheses == -1 )
         {
            ignoreParenthesis = FALSE;
            parentheses = 0;
            RETURN_TOKEN( TOKEN_TEXT );
         }

         mismatchParentheses = TRUE;
         LEXICAL_ERROR_MSG( "parenthesis mismatch" );
         RETURN_TOKEN_ERROR();
      }

     /*
      * If a '(' of a cast was parsed, does this ')' match it?
      */
      if ( cast  &&  (parentheses+1 == cast) )
         { cast = 0; RETURN_TOKEN( TOKEN_TEXT ); }

      if ( parentheses == 0  ||  parsingCallParms )
      { RETURN_TOKEN( ')' ); }
      else
      { RETURN_TOKEN( TOKEN_TEXT ); }
   }

  /*
   * To simplify the parser rules, ignore nested braces.
   * (Since any brace not on beginning of line is filtered-out,
   *  nested braces are less of a problem.)
   */
   if ( token == '{' )
   {
      ++braces;
      if ( braces < 0 )
      {
         mismatchBraces = TRUE;
         LEXICAL_ERROR_MSG( "curly brace '{' mismatch" );
         RETURN_TOKEN_ERROR();
      }

      if ( braces == 1 )
      { RETURN_TOKEN( '{' ); }
      else
#if 0
      { RETURN_TOKEN( TOKEN_TEXT ); }
#else
      /*
       * A func def might have more than one beginning '{' if
       * a #ifdef conditionally compiles different parameters types.
       * Try to ignore a superfluous '{'.
       */
      { braces = 1; RETURN_TOKEN( TOKEN_TEXT ); }
#endif
   }
   else if ( token == '}' )
   {
      --braces;
      if ( braces < 0 )
      {
         mismatchBraces = TRUE;
         LEXICAL_ERROR_MSG( "curly brace '}' mismatch" );
         RETURN_TOKEN_ERROR();
      }

      if ( braces == 0 )
      { RETURN_TOKEN( '}' ) }
      else
      { RETURN_TOKEN( TOKEN_TEXT ) }
   }

  /*
   * Return numeric value of token to yacc.
   */
   RETURN_TOKEN( token );
}

/*****************************************************************************
 * Must be called before lexing a new buffer.
 *****************************************************************************/
static void
LexReset( void )
{
   lineLex				= 1;
   discard				= 0;
   parentheses			= 0;
   braces				= 0;
   prevToken			= 0;
   prevTokenPassed		= 0;
   cast					= 0;
   mismatchBraces		= 0;
   mismatchParentheses	= 0;
   ignoreParenthesis	= 0;
}

/*****************************************************************************
 * Recover if any error happened (can be called if there was no error).
 *****************************************************************************/
static void
LexErrorRecover( void )
{
   if ( mismatchParentheses ) { ignoreParenthesis = 0;
                                cast = 0;
                                mismatchParentheses = parentheses = prevToken = prevTokenPassed = 0; }
   if ( mismatchBraces      ) { mismatchBraces      = braces      = prevToken = prevTokenPassed = 0; }
}

/*****************************************************************************
 * What to do if end of input is reached.
 *****************************************************************************/
int
yywrap()
{
   if ( parserMode == PARSER_MODE_HYPERSRC )
   {
#if 1
      return 1;
#else
      yyterminate();
#endif
   }

   exit(0);
   return 0; /* shutup compiler */
}
