/*****************************************************************************
 * $Id: parser.y,v 1.17 2001/03/04 18:22:51 jimb Exp $
 * Author		: Jim Brooks	<jim@jimbrooks.org>
 * Description	: Parses function definitions and function calls in C/C++ code.
 *                This parser isn't perfect.  It might miss a few function calls,
 *                and misjudge some lines as parse errors where a compiler wouldn't.
 *
 * Notes        : Note that function calls within a function definition are
 *                reported before the function definition itself.
 *                The reason is the token for '}' is parsed last (last line #).
 *                However, before '}' is parsed, the id (name) of the func def
 *                has been parsed and recorded.
 *
 *                Eg:
 *                LoadFile{} calls fclose()  line # 67 
 *                LoadFile{} calls g_free()  line # 82 
 *                LoadFile{} calls Popup()  line # 83 
 *                LoadFile{} body resides at lines 32..85 
 *****************************************************************************/


/*
 * token struct.
 */
%union
{
   char* str;
}

/*
 * tokens
 */
%nonassoc TOKEN_ID
%nonassoc TOKEN_TEXT
%nonassoc TOKEN_ERROR
%nonassoc TOKEN_IGNORE
%nonassoc TOKEN_KLUDGE_FUNC_PROTO_END


/*===========================================================================*/


%{

#define PARSER_Y

#if 0
#define DEBUG
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "parser.h"
#include "str_heap.h"
#include "treefunc.h"

int yyerror( );

#define MAX_ID_LEN	256

enum
{
   RESET_FULL,		/* full reset */
   RESET_STMT,		/* reset parser after a statement is sucessfully parsed */
   RESET_ERROR		/* reset parser after a parse error occurs */
};

STR_HEAP strHeap = { NULL, 0, NULL };
static char funcDefId[MAX_ID_LEN];
static char funcCallId[MAX_ID_LEN];
static char id[MAX_ID_LEN];
       char parserInputName[MAX_ID_LEN];

int parserMode = PARSER_MODE_PRINTF;

/*
 * Vars that should be initialized once every time ParserCreate() is called.
 */
static int initialized			= FALSE;
static int lineLeftParenthesis	= 0;
static int lineLeftBrace		= 0;
static int lineRightBrace		= 0;

/*
 * Vars that should be reset whenever ParserCreate() is called
 * and/or a parse error occurs.
 */
int parsingFuncDef				= 0;
int parsedLeftParOfFuncDef		= 0;
int parsingCallParms			= 0;
int betweenParmsAndBody			= 0;

extern char* yytext;
extern int yylex();
extern int lineLex;

/*****************************************************************************
 * For non-parse internal errors caused bugs, insufficent memory, bad pointers, etc.
 *****************************************************************************/
#define YACC_INTERNAL_ERROR_MSG( str ) \
{ \
   printf( "yacc INTERNAL ERROR: %s \n", str ); \
}

/*****************************************************************************
 * For parse errors present in the input being parsed.
 *****************************************************************************/
void PARSE_ERROR_MSG( char* str )
{
#ifdef LEX_YACC_QUIET
   return;
#endif

   if ( parserInputName[0] )
      printf( "%s:[%d]", parserInputName, lineLex );
   else
      printf( "[%d]", lineLex );

   printf( " parse error" );

   if ( str[0] ) printf( ": %s", str ); 

   printf( "\n" );
}

/*****************************************************************************
 * What to do if an internal error happens.
 *****************************************************************************/
#define YACC_ABORT() exit(1);

/*****************************************************************************
 * Record a parsed identifier.
 * No more than "destSize" chars will be copied.
 *****************************************************************************/
static void
RecordId( char* dest, char* src, int destSize  )
{
#ifdef HYPERSRC_PEDANTIC
  /*
   * Check before copying the string.
   */
   if ( src == NULL  ||  strlen( src ) > destSize )
   {
      if      ( dest == funcDefId )
         { YACC_INTERNAL_ERROR_MSG( "## Id of a func def " ); }
      else if ( dest == funcDefId )
         { YACC_INTERNAL_ERROR_MSG( "## Id of a func call " ); }
      else
         { YACC_INTERNAL_ERROR_MSG( "## Id of something " ); }

      if ( src == NULL )
         { YACC_INTERNAL_ERROR_MSG( "has a NULL ptr." ); }
      else
         { YACC_INTERNAL_ERROR_MSG( "is a string that is too large." ); }

      YACC_INTERNAL_ERROR_MSG( "\n## Aborting. \n" );
      YACC_ABORT();
   }
#endif /* HYPERSRC_PEDANTIC */

  /*
   * Safe to record id.
   */
   strncpy( dest, src, destSize );

   return;
}

/*****************************************************************************
 * Called when a function definition is completely parsed (ie '}' is encountered).
 *****************************************************************************/
static void
Yacc_stmt_func_def( void )
{
   switch ( parserMode )
   {
      case PARSER_MODE_HYPERSRC:
         AddFuncDefToGraph( funcDefId, lineLeftParenthesis, lineRightBrace );
         return;

      case PARSER_MODE_PRINTF:
         printf( "* %s{}", funcDefId );
#ifdef DEBUG
         printf( "    (%d..%d)", lineLeftParenthesis, lineRightBrace );
#endif
         printf( "\n\n\n" );
         return;
   }
}

/*****************************************************************************
 * Called when an identifier of a function call is parsed.
 *****************************************************************************/
static void
Yacc_func_call_id( char* pFuncCallId )
{
  /*
   * Ignore if id is really a C keyword that is followed by parentheses,
   * rather than a function call.
   */
#if 0
/* ----- to speed startup, this fixup is deferred to functree.c ----- */
   if (    strcmp( pFuncCallId, "if"      ) == 0
        || strcmp( pFuncCallId, "for"     ) == 0
        || strcmp( pFuncCallId, "while"   ) == 0
        || strcmp( pFuncCallId, "switch"  ) == 0
        || strcmp( pFuncCallId, "return"  ) == 0
        || strcmp( pFuncCallId, "sizeof"  ) == 0
        || strcmp( pFuncCallId, "defined" ) == 0 )
      return;

   if (    strcmp( funcDefId,   "if"      ) == 0
        || strcmp( funcDefId,   "for"     ) == 0
        || strcmp( funcDefId,   "switch"  ) == 0
        || strcmp( funcDefId,   "while"   ) == 0
        || strcmp( funcDefId,   "return"  ) == 0
        || strcmp( funcDefId,   "sizeof"  ) == 0
        || strcmp( funcDefId,   "defined" ) == 0 )
      return;
#endif

  /*
   * Record id of function call.
   */
   RecordId( funcCallId, pFuncCallId, sizeof(funcCallId) );

  /*
   * Report this function call.
   */
   switch ( parserMode )
   {
      case PARSER_MODE_HYPERSRC:
         AddFuncCallToGraph( funcCallId, lineLex );
         return;

      case PARSER_MODE_PRINTF:
#if 0
         /* try to resemble ctags style */
         printf( "%s() in %s{}   call  %d   \n", funcCallId, funcDefId, 0 );
#else
         /* more readable */
         printf( "    +--> %s() called", funcCallId );
#ifdef DEBUG
         printf( "    (%d)", lineLex );
#endif /*DEBUG*/
         printf( "\n" );
#endif
         return;
   }
}

%}

%%


/*===========================================================================*/


/*****************************************************************************
 * This set of context/tricks/kludges are employed to simplify/reduce the yacc rules
 * which otherwise would have been far more complex and difficult to write.
 *
 * - a parenthesis context: lex really returns nested parentheses  as TEXT tokens
 *                          unparenthesized commas returned as TEXT tokens
 * - nested parentheses   : lex only passes nested parentheses when the parser
 *                          is parsing the parameter list of a function call.
 * - a curly brace context: lex really returns nested curly braces as TEXT tokens
 * - a C comment   context: lex returns everything comment as TEXT tokens
 * - type casts           : lex filters out type casts
 * - "if ("               : lex filters out the parenthesis so that func calls
 *                          within a if() will be parsed (usually nested parentheses
 *                          are ignored).
 * - ");"                 : lex identifies this pair as the end of a func prototype,
 *                          otherwise protos will be mistaken by parser as func defs.
 *
 * Notes: The "error" token should be placed in a rule that is closest
 *        to where an error might occur, so that error recovery will
 *        occur sooner.  Writing the "error" token in the "program" or "stmt_s"
 *        rules would cause a lot of input to be skipped while recovering.
 *****************************************************************************/

program:
		stmt_s
	|	TOKEN_ERROR
	|	TOKEN_IGNORE
	;

stmt_s:
		stmt
	|	stmt_s   stmt
	;

stmt:
		stmt_func_def
	|   text_id_s   '{'   text_id_s   '}'
	;

/*****************************************************************************
 * function definition rules
 *
 * text_id_s_or_nothing accounts for old C style code in which parm types
 * were written between ')' and '{'.
 *****************************************************************************/

stmt_func_def:
		stmt_func_def_id   stmt_func_def_left_par   stmt_func_def_parm_s_or_nothing   stmt_func_def_right_par   text_id_s_or_nothing   stmt_func_def_left_brace   stmt_func_body_or_nothing   stmt_func_def_right_brace
								{ ParserReset( RESET_STMT ); Yacc_stmt_func_def(); }
	|	stmt_func_def_id   stmt_func_def_left_par   stmt_func_def_parm_s_or_nothing   stmt_func_def_right_par   TOKEN_KLUDGE_FUNC_PROTO_END
								{ ParserReset( RESET_STMT ); } /* a func prototype was parsed */
	|   error   TOKEN_TEXT		{ ParserErrorRecover(); yyerrok; }
	|   error   TOKEN_ID		{ ParserErrorRecover(); yyerrok; }
	;

stmt_func_def_left_par:
		'('						{ lineLeftParenthesis = lineLex;
                                  parsedLeftParOfFuncDef = 1; }
	;

stmt_func_def_right_par:
		')'						{ betweenParmsAndBody = 1; } /* parser is between list of parms and body of func def */
	;

stmt_func_def_left_brace:
		'{'						{ betweenParmsAndBody = 0;
                                  lineLeftBrace = lineLex; }
	;

stmt_func_def_right_brace:
		'}'						{ lineRightBrace = lineLex; }
	;

stmt_func_def_id:
		text_id_s				{ parsingFuncDef = 1; RecordId( funcDefId, id, sizeof(funcDefId) ); }
	;

stmt_func_def_parm_s_or_nothing:
		{ /* nothing */ }
	|	stmt_func_def_parm_s
	;

stmt_func_def_parm_s:
		text_id_s
	|	stmt_func_def_parm_s   ','   text_id_s
	;

stmt_func_body_or_nothing:
		{ /* nothing */ }
	|	stmt_func_body
	;

stmt_func_body:
		stmt_func_body_
	|	stmt_func_body   stmt_func_body_
	;

stmt_func_body_:
		text_id
	|	func_call
	;

/*****************************************************************************
 * function call rules
 *
 * Notes: Conditional compilation can cause a parse error in the func_call* rules.
 *        Eg:
 *              foo(
 *        #if 0
 *              a, b);
 *        #else
 *              a, b, c);
 *        #endif
 *
 *        "yyvalstr" is a kludge to suppress gcc warning that $<str>$
 *        is uninitialized (warning appears when using gcc's -O3/-inline).
 *****************************************************************************/

func_call:
		func_call_id   func_call_left_par   func_call_parm_s_list_or_nothing   func_call_right_par
	|   error   TOKEN_TEXT		{ yyerrok; }
	|   error   TOKEN_ID		{ yyerrok; }
	;

func_call_id:
		TOKEN_ID				{ char* yyvalstr = $<str>$; Yacc_func_call_id( yyvalstr ); }
	;

func_call_left_par:
		'('						{ ++parsingCallParms; }
	;

func_call_right_par:
		')'						{ --parsingCallParms; }
	;

func_call_parm_s_list_or_nothing:
		{ /* nothing */ }
	|	func_call_parm_s_list
	;

func_call_parm_s_list:
		func_call_parm_s
	|	func_call_parm_s_list   ','   func_call_parm_s
	;

func_call_parm_s:
		func_call_parm
	|	func_call_parm_s   func_call_parm
	;

func_call_parm:
		func_call
	|	TOKEN_TEXT
	|	TOKEN_ID
	;

/*****************************************************************************
 * misc rules
 *****************************************************************************/

text_id_s_or_nothing:
		{ /* nothing */ }
	|	text_id_s
	;

text_id_s:
		text_id
	|	text_id_s   text_id
	;

text_id:
		TOKEN_TEXT				{ char* yyvalstr = $<str>$; RecordId( id, yyvalstr, sizeof(id) ); }
	|	TOKEN_ID				{ char* yyvalstr = $<str>$; RecordId( id, yyvalstr, sizeof(id) ); }
	;

%%


/*===========================================================================*/


/*----------------------------------------------------------------------------
 * public functions
 *---------------------------------------------------------------------------*/


/*****************************************************************************
 * For hypersrc, parse functions in a module (a source file).
 *****************************************************************************/
void
ParseFuncsInModule( char* pInput, long inputSize, char* pInputName, int debug )
{
   void*	hnd = NULL;

   if ( pInputName )
      strncpy( parserInputName, pInputName, sizeof(parserInputName) );
   else
      parserInputName[0] = 0;

#ifdef HYPERSRC_PEDANTIC
   if ( pInput == NULL ) { YACC_INTERNAL_ERROR_MSG( "ParseFuncsInModule() pInput==NULL." ); YACC_ABORT(); }
#endif

   ParserCreate( debug ? PARSER_MODE_PRINTF : PARSER_MODE_HYPERSRC,
                 inputSize * 2 /* extra to be safe */ );
   hnd = LexBufferUse( pInput );
      yyparse();
   LexBufferDiscard( hnd );
   ParserDelete();

   return;
}


/*----------------------------------------------------------------------------
 * internal functions
 *---------------------------------------------------------------------------*/


/*****************************************************************************
 * Initialize the parser.
 *
 * Parms:		mode
 *				The parser's mode: one of PARSER_MODE_*.
 *
 *				strHeapSize
 *				The size of string heap to be allocated.
 *              A safe size would be double the size of the input to be parsed.
 *****************************************************************************/
static void
ParserCreate( int mode, int strHeapSize )
{
#ifdef HYPERSRC_PEDANTIC
   if ( initialized ) { YACC_INTERNAL_ERROR_MSG( "ParserCreate() can't be nested." ); YACC_ABORT(); }
#endif

  /*
   * Reset state.
   */
   ParserReset( RESET_FULL );

  /*
   * Record parser's mode.
   */
   parserMode = mode;

  /*
   * Create a heap to hold strings that will be parsed.
   */
   StrHeapCreate( &strHeap, strHeapSize );

   return;
}

/*****************************************************************************
 * Delete resources used by parser.
 *****************************************************************************/
static void
ParserDelete( void )
{
   initialized = FALSE;

   StrHeapDelete( &strHeap );
}

/*****************************************************************************
 * Reset parser.
 * Pass RESET_*.
 *****************************************************************************/
static void
ParserReset( int howMuchToReset )
{
   switch ( howMuchToReset )
   {
      case RESET_FULL:
      initialized			= TRUE;
      lineLeftParenthesis	= 0;
      lineLeftBrace			= 0;
      lineRightBrace		= 0;

      /* -- fall thru -- */

      case RESET_STMT: case RESET_ERROR: default:
      parsingFuncDef		= 0;
      parsedLeftParOfFuncDef = 0;
      parsingCallParms		= 0;
      betweenParmsAndBody	= 0;
   }
}

/*****************************************************************************
 * Recover from a parse error.
 *****************************************************************************/
static void
ParserErrorRecover( void )
{
   ParserReset( RESET_ERROR );
}

/*****************************************************************************
 * What to do if a parse error is detected in the input.
 *****************************************************************************/
int
yyerror( char* s )
{
   PARSE_ERROR_MSG( "" );
   return 0; /* shutup compiler */
}

/*****************************************************************************
 * For testing.
 *****************************************************************************/
#if 0
#ifdef DEBUG
static char*
TestCase( void )
{
   char* pInput = (char*)malloc( 256 );

   strcpy( pInput, " void foo( int a ) { xxxx } \n" );
   strcpy( pInput, " unsigned int foo( volatile int a , signed char c ) {  ff();  } \n" );
   strcpy( pInput, " int foo( int a ) {  f(); fab ( a, b );  } \n" );
   strcpy( pInput, " int foo( int a )\n{\n\tf(); foo( a, b ); bar(); *p(); pf->virtual();  } \n" );
   strcpy( pInput, " int foo( int a )\n{\n   bar();\n}\nvoid foo2( char c)\n{\n\tbar2( 1 );\n}\n" );
   strcpy( pInput, "int dummy( int a )\n{\n   f();\n}" );

   return pInput;
}
#endif /* DEBUG */
#endif
