/******************************************************************************
 * $Id: tags.c,v 1.158 2001/06/19 01:45:36 jimb Exp $
 * Author		: Jim Brooks	<jim@jimbrooks.org>
 * Description	: A set of functions for managing source code tags.
 ******************************************************************************/

#include <unistd.h>

#include "header.h"

/*
 * Local structs/vars.
 */
static GList* pGlistLastTag;		/* points to last link of tags glist,          updated by AddCtagsLineToTagsGlist() */
static GList* pGlistLastFuncTag;	/* points to last link of function tags glist, updated by AddCtagsLineToTagsGlist() */

/*
 * For a narrower list column, some ctag types are abbreviated.
 */
static int*	validTagTypesLen;		/* table of lenths, built at run-time */
const char* validTagTypes[] =
{
  /*
   * C tag types:
   */
   "func",
   "var",
   "macro",
   "member",
   "enum",
   "typedef",
   "struct",
   "union",
   "proto",
#if 0
   "define",			/* ctags outputs "macro" instead */
   "externvar",			/* by default, ctags won't produce this */
#endif

  /*
   * C++ tag types:
   */
   "class",
   "namespace",

  /*
   * Assembly tag types:
   */
   "label",

  /*
   * M4 tag types:
   */
   "m4_macro",

  /*
   * Is this a non-tag?
   */
   "(NONE)"
};

/*****************************************************************************
 * Prepare state for process ctags output.
 *****************************************************************************/
void
PrepareProcessCtags( void )
{
   int			i;

  /*
   * Init tags global vars.
   */
   tagsCnt = 0;
   ctagsLines = 0;

  /*
   * Build table of lengths of valid tag types (to speed CheckTagType() ).
   */
   validTagTypesLen = g_malloc( ELEMS(validTagTypes) * sizeof(int) );
   for ( i = 0; i < ELEMS(validTagTypes); i++ )
   { validTagTypesLen[i] = strlen( validTagTypes[i] ); }

   return;
}

/*****************************************************************************
 * Process ctags output from a file, resulting in a GList of TAG structs.
 *
 * Parms   : pGlistTags (OUT/GLOBAL)
 *           GList of TAG structs.
 *
 *           tagsCnt (OUT/GLOBAL)
 *           Count of links in pGlistTags.
 *****************************************************************************/
void
ProcessCtagsFromFile( char* pFilename )
{
   char*	pCtags;
   long		ctagsSize;

  /*
   * Were there too many tags last time?
   */
   if ( tooManyTags ) return;

  /*
   * Prepare state.
   */
   PrepareProcessCtags();

  /*
   * Create a list of tags from ctags.
   */
   pCtags = ReadCtagsFromFile( &ctagsSize, /* out */
                               pTagsFilename );
   pGlistTags = ProcessCtagsBuf( pCtags, ctagsSize );
   if ( tooManyTags ) WarnTooManyTags();

   return;
}

/*****************************************************************************
 * Process ctags output from stdin, resulting in a GList of TAG structs.
 *
 * Parms   : pGlistTags (OUT/GLOBAL)
 *           GList of TAG structs.
 *
 *           tagsCnt (OUT/GLOBAL)
 *           Count of links in pGlistTags.
 *****************************************************************************/
void
ProcessCtagsFromStdin( void )
{
   char*	pBuf;
   int      bytesRead;
   int      bytesReadMore;
   GList*	pGlistPiece = NULL;

  /*
   * Prepare state.
   */
   PrepareProcessCtags();

  /*
   * Allocate enough memory for a buffer that might be > BUFSIZ.
   */
   pBuf = g_malloc( BUFSIZ * 10 );
   g_assert( pBuf );

  /*
   * This loop grabs pieces of ctags output, from stdin, and passes each piece
   * to ProcessCtagsBuf().  Each partial GList is appended to what will be
   * the final GList.
   */
   for (;;)
   {
     /*
      * Read the next BUFSIZ piece of ctags lines from stdin.
      * Break if end of stdin.
      */
      bytesRead = read( STDIN_FILENO, pBuf, BUFSIZ );
      if ( bytesRead <= 0 ) break;

     /*
      * Rarely, if only one byte has been read, then read a second byte,
      * to let the following loop assume that at least two bytes were read.
      */
      if ( bytesRead == 1 )
      {
         bytesReadMore = read( STDIN_FILENO, &pBuf[bytesRead], 1 );
         if ( bytesReadMore <= 0 ) goto skip;
         ++bytesRead;
      }

     /*
      * Ctags lines can not be passed truncated, so unless no more input,
      * and if the last char(s) weren't EOL, try to read more chars until
      * an EOL is encountered.
      */
      while (    ! IsEolUnix(  &pBuf[bytesRead-1] )
              && ! IsEolMsdos( &pBuf[bytesRead-2] ) ) /* <-- assumes at least 2 bytes were read */
      {
         bytesReadMore = read( STDIN_FILENO, &pBuf[bytesRead], 1 );
         if ( bytesReadMore <= 0 ) break;
         ++bytesRead;
      }

      skip:

     /*
      * Truncate buffer (though not strictly necessary).
      */
      pBuf[bytesRead] = 0;

     /*
      * Process this piece (which has untruncated lines).
      * If this is not the first piece, then append this GList into the final GList.
      */
      if ( pGlistTags == NULL )
      {
         pGlistTags = ProcessCtagsBuf( pBuf, bytesRead );
         if ( pGlistTags == NULL )
            Error( "Could not process any ctags lines, bailing out!" );
      }
      else
      {
         pGlistPiece = ProcessCtagsBuf( pBuf, bytesRead );
         if ( pGlistPiece == NULL )
            break;
         g_list_concat( pGlistTags, pGlistPiece );
      }

     /*
      * Stop if there were too many tags.
      */
      if ( tooManyTags )
      {
         WarnTooManyTags();
         break;
      }

     /*
      * Continue loop.
      */
   }

  /*
   * Were any ctag lines processed?
   */
   if ( pGlistTags == NULL )
       Error( "No ctags lines were processed, bailing out!" );

   g_free( pBuf );

   return;
}

/******************************************************************************
 * Process a buffer of lines from ctags.
 *
 * - Counts each ctags line (++ctagsLine)
 * - Creates a glist of TAG structs.
 * - Creates a hash table containing info about each module.
 *   This hash table is the first level of the func graph.
 * - Stops if tag count exceeds a limit.
 *
 * Returns : Returns a GList.
 *
 * Parms   : pCtags
 *           Pointer to raw ctags output.
 *
 *           ctagsSize
 *           Size of ctags output.
 *
 *           pHashModules (GLOBAL)
 *
 *           tooManyTags (OUT/GLOBAL)
 *           Will be TRUE if this function stopped
 *           when there were too many tags.
 *
 *           ctagsLine (OUT/GLOBAL)
 *           Incremented.
 ******************************************************************************/
GList*
ProcessCtagsBuf( char* pCtags,
                 long  ctagsSize )
{
   const char*	pLine		= pCtags;
   GList*		pGlist		= NULL;
   struct TAG*	pTag		= NULL;		/* threading precludes making these static! */
   struct TAG*	pTagPrev	= NULL;

  /*
   * Ensure the tags glist pointer is NULL.
   * (The GLIB list functions will create a list if the passed list pointer is NULL.)
   */
   pGlist = NULL;

  /*
   * Decompose each ctags line.
   * Add each decomposed part to a TAG struct.
   */
   for (;;)
   {
     /*
      * Count this ctags line, and stop if there were too many tags.
      */
      ++ctagsLines;
      if ( ctagsLines > maxTags )
      {
         tooManyTags = TRUE;
         break;
      }

     /*
      * Is this ctags line empty? (very unexpected)
      */
      if ( IsLineEmpty( pLine ) )
         break;

     /*
      * Add this ctags line to a glist.
      * Later, the glist will be used to create a list widget.
      */
      pTagPrev = pTag;
      if ( AddCtagsLineToTagsGlist( pLine, &pGlist, &pGlistLastTag, &pTag ) == FUNC_OK )
      {
         P_HASH_ITEM_MODULE	pHashItem;

        /*
         * Insert/replace this module into a hash table.
         * Unless the module name hasn't changed in the last two ctags lines.
         */
         if (    pTagPrev == NULL
              || CompareStrings( pTagPrev->pModule, pTag->pModule ) != 0 )
         {
           /*
            * The hash item already exists if first line of this buffer
            * and the last line of the previous buffer are all part of the same module.
            */
            if ( g_hash_table_lookup( pHashModules, pTag->pModule ) == NULL )
            {
               StatusMsg( "Processing module: %s \n", pTag->pModule );

              /*
               * For this newly-encountered module (referenced by a field in a ctags line),
               * create a hash item, and insert/replace it into a hash table.
               */
               pHashItem = (P_HASH_ITEM_MODULE) g_malloc( sizeof(HASH_ITEM_MODULE) );

               pHashItem->pModuleName    = CloneString( pTag->pModule );
               pHashItem->pModuleBuffer  = NULL;			/* not loaded */
               pHashItem->moduleSize	  = 0;
               pHashItem->pGslistFuncDef = NULL;			/* to assign after all tags are processed */
               pHashItem->pGlistTag      = pGlistLastTag;	/* ##!! INVALIDATED when tags are sorted !!## */

               g_hash_table_insert( pHashModules,
                                    (gpointer) pTag->pModule,
                                    (gpointer) pHashItem );
            }
         }
      }
      else
      {
        /*
         * Oops, unable to process this ctags line.
         * This is considered a recoverable error (albeit unexpected)
         * only if previous ctags lines have been processed successfully,
         * which is indicated if the list does contains some elements.
         */
         if ( pGlist == NULL )
            Error( "Could not process any ctags lines, bailing out!" );

         SetStatusbarInitialMsg( "Could not process a ctags line, proceeding anyway." );
      }

     /*
      * Move to next line in ctags.
      */
      pLine = NextLine( pLine, MAX_CTAGS_LINE_LEN );

     /*
      * Break if no more lines in ctags.
      */
      if (    pLine == NULL
           || (pLine - pCtags) >= ctagsSize ) break;
   }

#if 0
   DebugMsg( "tag lines processed = %d", ctagsLines );
#endif

   return pGlist;
}

/*****************************************************************************
 * Warn (only once) that there were too many tags to process.
 *****************************************************************************/
void
WarnTooManyTags( void )
{
   static gboolean warned = FALSE;
   static char str[] = "There were too many tags, stopped tag processing (use -max-tags to override).";

   if ( ! warned )
   {
      warned = TRUE;
      SetStatusbarInitialMsg( str );
      Warning( str );
   }
   return;
}

/******************************************************************************
 * Resort the tag glist according to one of: tag name, tag type, or module.
 *
 * The resort is accomplished by creating a new glist in order to use
 * g_list_insert_sort() to do the sorting work.
 *
 * Parms   : sort
 *           One of SORT_*.
 *
 *         : pGlistTags (IN/OUT/GLOBAL)
 *           The current tags glist will be freed and this will be repointed
 *           to the new sorted glist.
 *
 *         : sort1 (OUT/GLOBAL)
 *
 *         : sort2 (OUT/GLOBAL)
 *
 *         : sort3 (OUT/GLOBAL)
 ******************************************************************************/
void
ResortTagsGlist( int sort )
{
  /*
   * Assign the sort vars appropriately.
   */
   switch ( sort )
   {
      case SORT_TAG:
         sort1    = SORT_TAG;
         sort2    = SORT_TYPE;
         sort3    = SORT_MODULE;
         break;

      case SORT_TYPE:
         sort1    = SORT_TYPE;
         sort2    = SORT_TAG;
         sort3    = SORT_MODULE;
         break;

      case SORT_MODULE:
         sort1    = SORT_MODULE;
         sort2    = SORT_TYPE;
         sort3    = SORT_TAG;
         break;

      default:
         g_assert( 0 );
   }

  /*
   * Prepare for displaying the completion percentage of sorted tags
   * in the statusbar.
   */
   tagsSorted			  = 0;
   tagsSortedPercentShown = 0;

  /*
   * Sort the tags.
   */
   pGlistTags = InsertionSortGlist( pGlistTags );

   MostlyClearStatusbar();

   return;
}

/******************************************************************************
 * Compare two TAG structs, using up to 3 levels of comparisons.
 *
 * Parms   : sort1 (GLOBAL)
 *
 *         : sort2 (GLOBAL)
 *
 *         : sort3 (GLOBAL)
 ******************************************************************************/
gint
ResortTagsGlistCompare( gconstpointer pTag1,
                        gconstpointer pTag2 )
{
   gint		rv;

  /*
   * Try the first-level sort.  Return if the tag fields are inequal.
   */
   rv = CompareTags( (const PTAG)pTag1,
                     (const PTAG)pTag2,
                     sort1 );
   if ( rv ) return rv;

  /*
   * Try the second-level sort.  Return if the tag fields are inequal.
   */
   rv = CompareTags( (const PTAG)pTag1,
                     (const PTAG)pTag2,
                     sort2 );
   if ( rv ) return rv;

  /*
   * Last, try the third-level sort.
   */
   rv = CompareTags( (const PTAG)pTag1,
                     (const PTAG)pTag2,
                     sort3 );
   return rv;
}

/******************************************************************************
 * Compare the members of two TAG structs.
 *
 * Returns : 0 if equal
 *
 * Parms   : pTag1
 *
 *         : pTag2
 *
 *         : member
 *           Which member of the TAG struct, use one of SORT_*.
 ******************************************************************************/
inline gint
CompareTags( const PTAG pTag1,
             const PTAG pTag2,
             int        member )
{
   switch ( member )
   {
      case SORT_TAG:
         return CompareStrings( pTag1->pName,
                                pTag2->pName );

      case SORT_TYPE:
         return CompareStrings( pTag1->pType,
                                pTag2->pType );

      case SORT_MODULE:
         return CompareStrings( pTag1->pModule,
                                pTag2->pModule );

      default:
         g_assert( 0 );
   }

   return 0; /* shutup compiler */
}

/******************************************************************************
 * Derive a TAG struct from a ctags line.
 * Add the TAG struct to a list.
 *
 * Parms   : pLine
 *           Pointer to start of a line in ctags.
 *
 *         : ppGList (OUT)
 *           Pointer to a GList pointer that is a list of TAG structs.
 *
 *         : ppGListLastLink (IN/OUT)
 *           Pointer to last link of tag glist (of recently-added tag).
 *           Will be updated as a new link will be appended.
 *
 *         : ppTag (OUT)
 *           Pointer to the resulting TAG struct created from ctags line.
 *
 * Returns : FUNC_OK or FUNC_FAIL
 *
 * This is an example of a line from "ctags -x":
 *
 *  tag:    type:       line:  module:     (ignored:)
 * "main    function    14     dummy.c     int main( int argc, char* argv[] )"
 ******************************************************************************/
FUNC_X
AddCtagsLineToTagsGlist( const char*  pLine,
                         GList**      ppGlist,         /*OUT*/
                         GList**      ppGlistLastLink, /*IN/OUT*/
                         struct TAG** ppTag            /*OUT*/ )
{
   struct TAG*	pTag	= NULL;
   const char*	pWord	= pLine;
   int			sizeWord;
   int          lineFieldRetries = 0;

  /*
   * Create a TAG struct (to be added to the list after it is successfully filled).
   */
   pTag = AllocTagStruct();
   *ppTag = pTag;

  /*
   * Process the tag name field.
   */
   sizeWord = ParseWord( pWord, &pWord, MAX_CTAGS_FIELD_LEN );
   if ( sizeWord > 0 )
   {
      pTag->pName = CreateString( sizeWord );
      CopyChars( pTag->pName, pWord, sizeWord );
   }
   else
      return FUNC_FAIL;
   pWord = PastWord( pWord, sizeWord );

  /*
   * Process the type field.
   */
   sizeWord = ParseWord( pWord, &pWord, MAX_CTAGS_FIELD_LEN );
   if ( sizeWord > 0 )
   {
     /*
      * Try to recognize the tag type.
      */
      pTag->pType = (char*) CheckTagType( pWord );		/* returns ptr to static tag type */
      if ( pTag->pType == NULL )
      {
        /*
         * Tag was not immediately recognized.
         * Is this the case in which Exuberant ctags outputs a multi-word
         * tag name for a C++ operator?  If so, the ctags line will have the format,
         * "operator X function nnn", in which "X" can be one or more words.
         */
         if ( FindSubstr( pTag->pName, "operator" ) )
         {
           /*
            * Reassign the tag name with "operator X"
            * (ie the string that precedes the "function" field.)
            */
            DeleteString( pTag->pName );
            pWord = FindSubstr( pLine, "function" );		/* pWord -> "function" */
#ifdef HYPERSRC_PEDANTIC
            if ( pWord == NULL  ||  pWord > NextLine( pLine, MAX_CTAGS_LINE_LEN ) ) goto unrecognized_tag;
#else
            if ( pWord == NULL ) goto unrecognized_tag;
#endif
            pTag->pName = CreateString( pWord - pLine );
            CopyChars( pTag->pName, pLine, pWord - pLine );

           /*
            * For this C++ operator, assign the tag type field as "function".
            */
            pTag->pType = "func";

           /*
            * Prepare to advance pWord past the "function" field.
            */
            sizeWord = sizeof("function")-1;

           /*
            * Successfully parsed "operator X".
            */
         }
         else
         {
           /*
            * Tag is not "operator X" and is not recognized.
            * WARNING: Ignoring this error could result in a failure later,
            *          but acceptAllTags is necessary to browse non-C src.
            */
            if ( !acceptAllTags )
               goto unrecognized_tag;

           /*
            * Fill TAG.pType with unrecognized tag type.
            */
            pTag->pType = CreateString( sizeWord );
            CopyChars( pTag->pType, pWord, sizeWord );
         }
      }
   }
   else
      return FUNC_FAIL;
   pWord = PastWord( pWord, sizeWord );

  /*
   * Process the line number field.
   */
retry_line_field:
   sizeWord = ParseWord( pWord, &pWord, MAX_CTAGS_FIELD_LEN );
   if ( sizeWord > 0 )
   {
     /*
      * Copy the line number in its string form.
      */
      pTag->pLine = CreateString( sizeWord );
      CopyChars( pTag->pLine, pWord, sizeWord );

     /*
      * Try to convert this field into an integer.
      */
      pTag->lineInt = atoi( pWord );

     /*
      * The result should be > 0.
      */
      if ( pTag->lineInt == 0 )
      {
        /*
         * A new multiword tag type might have been encountered.
         * Retry the next few words.  Hopefully, one might be the line # field.
         * (By neglect, only the first word of such a tag type will be shown in list.)
         */
         if ( lineFieldRetries++ < 3 )
         {
            pWord = PastWord( pWord, sizeWord );
            goto retry_line_field;
         }
         return FUNC_FAIL;
      }
   }
   else
      return FUNC_FAIL;
   pWord = PastWord( pWord, sizeWord );

  /*
   * Process the module (ie file) field.
   * TODO: Does not account for filenames with spaces (though very rare).
   */
   sizeWord = ParseWord( pWord, &pWord, MAX_CTAGS_FIELD_LEN );
   if ( sizeWord > 0 )
   {
      pTag->pModule = CreateString( sizeWord );
      CopyChars( pTag->pModule, pWord, sizeWord );

     /*
      * Does this look like a module?
      */
      /* TODO */
   }
   else
      return FUNC_FAIL;
   pWord = PastWord( pWord, sizeWord );

  /*
   * Add the TAG struct to the list.
   */
   *ppGlist = fast_g_list_append( *ppGlist,
                                  (gpointer) pTag,
                                  ppGlistLastLink /*IN/OUT*/ );
#ifdef HYPERSRC_PEDANTIC
   g_return_val_if_fail( *ppGlist, FUNC_FAIL );
#endif

  /*
   * If this is the tag of a function, then also add it to a glist containing
   * only function tags (to speed FindFuncTag() while building func graph/tree).
   */
   if ( *(guint32*)pTag->pType == ('f'|'u'<<8|'n'<<16|'c'<<24) )
   {
      pGlistFuncTags = fast_g_list_append( pGlistFuncTags,
                                           (gpointer) pTag,
                                           &pGlistLastFuncTag /*IN/OUT*/ );
#ifdef HYPERSRC_PEDANTIC
      g_assert( pGlistFuncTags );
#endif
   }

  /*
   * Avoid duplicating strings in multiple tags.
   */
   if ( conserveMemory ) MinimizeTagStruct( *ppTag, *ppGlist );

  /*
   * Count this tag (so that the slow g_list_length() can be avoided).
   */
   ++tagsCnt;

   return FUNC_OK;

unrecognized_tag:
   if ( ExtractLine( pLine, MAX_CTAGS_LINE_LEN ) )
   { Error( "Couldn't recognize tag.\nTry passing -accept-all-tags (as a last resort).\nctags line:\n%s\n", ExtractLine( pLine, MAX_CTAGS_LINE_LEN ) ); }
   else
   { Error( "Couldn't recognize tag.\nTry passing -accept-all-tags (as a last resort)." ); }
   return FUNC_FAIL; /* shutup compiler */
}

/*****************************************************************************
 * Minimize the amount of memory needed for a tag struct.
 *****************************************************************************/
void
MinimizeTagStruct( struct TAG* pTag,
                   GList*      pGlist )
{
   P_HASH_ITEM_MODULE	pHashItem;
   GList*				pGlistTmp;

  /*
   * Would this duplicate a module name string already in the module hash table?
   */
   pHashItem = (P_HASH_ITEM_MODULE) g_hash_table_lookup( pHashModules, pTag->pModule );
   if ( pHashItem )
   {
      DeleteString( pTag->pModule );
      pTag->pModule = pHashItem->pModuleName;
   }

  /*
   * Would this duplicate a tag name string already in a prior tag struct?
   */
   for ( pGlistTmp = pGlist;
         (void*)pGlistTmp->data != (void*)pTag;		/* don't compare the tag against itself! */
         pGlistTmp = pGlistTmp->next )
   {
     /*
      * Found duplicate tag name string?
      */
      if ( CompareStrings( pTag->pName, ((struct TAG*)(pGlistTmp->data))->pName ) == 0 )
      {
         DeleteString( pTag->pName );
         pTag->pName = ((struct TAG*)pGlistTmp->data)->pName;
         break;
      }
   }

   return;
}

/******************************************************************************
 * Allocate a TAG struct.
 ******************************************************************************/
struct TAG*
AllocTagStruct( void )
{
  /*
   * The TAG struct is initially all clear.
   */
   return g_malloc0( sizeof(struct TAG) );
}

/******************************************************************************
 * Free a TAG struct.
 ******************************************************************************/
void
FreeTagStruct( struct TAG* pTag )
{
  /*
   * TODO
   */
}

/******************************************************************************
 * Assign the members of TAG struct.
 * Memory for each member will allocated here.
 ******************************************************************************/
void
FillTagStruct( struct TAG* pTag,
               gchar*      pName,
               gint        lineNum,
               gchar*      pType,
               gchar*      pModule )
{
   /* TAG.pName */
   pTag->pName = CreateString( StrLen(pName) );
   strcpy( pTag->pName, pName );

   /* TAG.pType */
   pTag->pType = CreateString( StrLen(pType) );
   strcpy( pTag->pType, pType );

   /* TAG.pLine */
   pTag->pLine = CreateString( 10 );
   sprintf( pTag->pLine, "%d", lineNum );

   /* TAG.lineInt */
   pTag->lineInt = lineNum;

   /* TAG.pModuleName */
   pTag->pModule = CreateString( StrLen(pModule) );
   strcpy( pTag->pModule, pModule );

   return;
}

/******************************************************************************
 * Read ctags -x output from a file.
 *
 * Returns : Returns pointer to ctags output (or abends).
 *
 * Parms   : pCtagsSize (out)
 *           Size of ctags output.
 *
 *           pFilename
 ******************************************************************************/
char*
ReadCtagsFromFile( long* pCtagsSize, /* out */
                   char* pFilename )
{
   gpointer	pBuffer;

   pBuffer =  LoadFile( pFilename,
                        pCtagsSize /* out */ );
   if ( pBuffer == NULL ) { Error( "\nCould not open tags file.\n\nTry running 'Hypersrc.pl' instead. \n" ); }

   return pBuffer;
}

/******************************************************************************
 * Ensure that a tag type (extracted from ctags) is recognizable.
 * 
 * Parms	: pTagType
 *            Pointer to tag type from ctags (doesn't need to be null-terminated)
 *
 * Returns  : If the tag type is valid, returns a pointer to a matching string
 *            IN STATIC MEMORY (to conserve memory by not allocating a new string).
 ******************************************************************************/
const char*
CheckTagType( const char* pTagType )
{
   int i;

#if 0
  /*
   * ## AddCtagsLineToTagsGlist() depends on CheckTagType()
   * ## returning failure when ctags outputs "operator X".
   * ## Otherwise, AddCtagsLineToTagsGlist() will malfunction.
   */
   if ( acceptAllTags ) return FUNC_OK;
#endif

   for ( i = 0; i < ELEMS(validTagTypes); i++ )
   {
      if ( memcmp( pTagType, validTagTypes[i], validTagTypesLen[i] ) == 0 )
         return validTagTypes[i];
   }

  /*
   * This tag type was not recognized.
   */
   return NULL;
}

/*****************************************************************************
 * Goto the tag which is under the cursor (of the text widget).
 *
 * Invoked when the toolbar button "Goto Tag" is selected.
 *****************************************************************************/
void
GotoTagUnderCursor( void )
{
   void GotoTagUnderCursor___( void );

  /*
   * For stability, serialize this function.
   */
   if ( busy_GotoTagUnderCursor ) { DebugMsg( "GotoTagUnderCursor() is busy." ); return; }
   busy_GotoTagUnderCursor = TRUE;
   GotoTagUnderCursor___();
   busy_GotoTagUnderCursor = FALSE;
}
void
GotoTagUnderCursor___( void )
{
   GtkWidget*	pWidgetTextActivePrev   = pWidgetTextActive;
   GString*		pGstr					= NULL;
   gchar*		pText					= NULL;
   guint		textLen;
   guint		cursorIdx;
   gchar		tagName[MAX_TAG_NAME_LEN];
   int			row;

  /*
   * Ensure the text widget is created and isn't empty.
   */
   if ( IsTextWidgetEmpty() )
   {
      PrintStatusbar( "Text widget is empty" );
      return;
   }

  /*
   * Obtain contents of text widget.
   */
   pText = ActiveTextWidgetContents( &textLen );
   if ( pText == NULL  ||  textLen > MAX_TEXT_BUFFER_LEN ) goto error;

  /*
   * Get index of cursor.
   */
   cursorIdx = TextCursorIdx();
   if ( cursorIdx < 0  ||  cursorIdx > textLen ) goto error;

  /*
   * Scoop word at cursor.
   * Return if cursor is not over a word.
   */
   pGstr = ScoopWordAtCursor( MAX_TAG_NAME_LEN,
                              pText,
                              textLen,
                              cursorIdx );
   if ( pGstr == NULL )
   {
      PrintStatusbar( "Please move the cursor over a word that might be a tag." );
      goto quit;
   }

   /*-----*/

   gtk_clist_freeze( GTK_CLIST( pWidgetClistTags ) );
   gtk_clist_freeze( GTK_CLIST( pWidgetClistModules ) );

  /*
   * Record the current cursor position into the tags history.
   * This tag is needed so that the user can return to previous cursor positions
   * (via Prev Tag button).
   */
   RecordCursorInTagHistory( "(recorded position)" );

  /*
   * Should the tagged line be shown in the main or aux text widget?
   */
   if ( ! auxKeyPressed )
      SetActiveTextWidget( pWidgetTextMain );
   else
      SetActiveTextWidget( pWidgetTextAux );

  /*
   * Select the row in the clist which contains the tag.
   * For SearchClistWidget(), create a string to hold the tag name.
   */
   if ( tagsCnt > TAGS_VERY_MANY ) PrintStatusbar( "Going to tag..." );

   CopyChars( tagName, pGstr->str, pGstr->len );
   row = \
   FindTagCheckMultipleMatches( tagName,
                                NULL,
                                widgetTagsClistSelectedRow,
                                TRUE,	/* be exact */
                                TRUE,   /* (case N/A) */
                                TRUE,   /* select notebook page */
                                "Multiple tags share the same name (press F5 for next one)." );

   if (    (tagsCnt > TAGS_VERY_MANY)
        && GetStatusbarMsg()
        && CompareStrings( GetStatusbarMsg(), "Going to tag..." ) == 0 )
      MostlyClearStatusbar();

   gtk_clist_thaw( GTK_CLIST( pWidgetClistTags ) );
   gtk_clist_thaw( GTK_CLIST( pWidgetClistModules ) );

  /*
   * Restore the active text widget.
   */
   pWidgetTextActive = pWidgetTextActivePrev;
   SetActiveTextWidget( pWidgetTextActive );

  /*
   * If the search failed then remove the tag of the current position
   * because since no hyperlink was done, the current position won't change.
   */
   if ( row == ROW_NONE)
   {
      UndoRecordCursorInTagHistory();
      PrintStatusbarThirds( "'", tagName, "' is not tagged" );
   }

  /*
   * Free mem, return.
   */
quit:
   if ( pGstr ) g_string_free( pGstr, TRUE );
   return;

  /*
   * Error occurred, but try to continue.
   */
error:
   WarningMalfunction();
   return;
}

/*****************************************************************************
 * Return true if a char can be part of a tag.
 *****************************************************************************/
int
IsTagChar( gchar c )
{
   return    isalnum( c )
          || (c == '_') ;
}

/*****************************************************************************
 * Find a tag.
 * Directly searches thru a glist.
 *
 * Parms	: pTagName
 *
 *            pTagType
 *            Pass NULL to match any.
 *****************************************************************************/
PTAG
FindTag( char* pTagName,
         char* pTagType )
{
   GList*	pGlist;

   for( pGlist = pGlistTags;
        pGlist;
        pGlist = pGlist->next )
   {
#ifdef HYPERSRC_PEDANTIC
g_assert( pGlist->data   &&  ((PTAG)pGlist->data)->pName  &&  ((PTAG)pGlist->data)->pType );
#endif

     /*
      * Compare tag name.
      */
      if ( STREQ( pTagName, ((PTAG)pGlist->data)->pName ) )
      {
        /*
         * Tag name matches.  Compare tag type (if specified).
         */
         if ( pTagType )
            if ( STRNE( pTagType, ((PTAG)pGlist->data)->pType ) )
               return NULL;

        /*
         * Return matching tag.
         */
         return (PTAG) pGlist->data;
      }
   }

   return NULL;
}

/*****************************************************************************
 * Find the tag of a function.
 * If the tag is sought in a particular module, instead call FindFuncTagInOneModule().
 *
 * Parms   : pFuncName
 *
 *           pModuleName
 *****************************************************************************/
PTAG
FindFuncTag( char*    pFuncName,
             char*    pModuleName )
{
   GList*	pGlist;
   PTAG		pTagAMatch = NULL;

  /*
   * Search thru the glist that only contains tags of functions.
   */
   for ( pGlist = pGlistFuncTags; pGlist; pGlist = pGlist->next )
   {
      PTAG	pTag = (PTAG) pGlist->data;

     /*
      * Does the function name match?
      */
      if ( CompareStrings( pFuncName, pTag->pName ) == 0 )
      {
        /*
         * Does the function reside in the preferred/required module?
         */
         if ( CompareStrings( pModuleName, pTag->pModule ) == 0 )
         {
           /*
            * Yes, this is the best match, return it.
            */
            return pTag;
         }

        /*
         * This function resides in another module.
         * Record it as a possible match.
         */
         pTagAMatch = pTag;
      }
   }

  /*
   * A preferred/required match in the named module wasn't found.
   * Return a less-preferrable match (if any).
   */
   return pTagAMatch;
}

/*****************************************************************************
 * Search thru one module to find the tag of a function.
 *
 * Parms   : pFuncName
 *
 *           pHashItem
 *           Which module to search.
 *
 * Assumes : Assumes the tags glist is newly created and is unsorted,
 *           because HASH_ITEM_MODULE->pTagFirst is relied upon.
 *****************************************************************************/
PTAG
FindFuncTagInOneModule( char*              pFuncName,
                        P_HASH_ITEM_MODULE pHashItem )
{
   GList*	pGlist;
   char*    pModuleNameToSearch;

   pModuleNameToSearch = pHashItem->pModuleName;

   for ( pGlist = pHashItem->pGlistTag; pGlist; pGlist = pGlist->next )
   {
      PTAG	pTag = (PTAG) pGlist->data;

     /*
      * Overrun into the next module?
      */
      if ( CompareStrings( pModuleNameToSearch, pTag->pModule ) )
         return NULL;

     /*
      * Is this the tag of a function?
      */
      if ( *(guint32*)pTag->pType == ('f'|'u'<<8|'n'<<16|'c'<<24) )
      {
        /*
         * This is a tagged function.  Does it name match?
         */
         if ( CompareStrings( pFuncName, pTag->pName ) == 0 )
            return pTag;
      }
   }

  /*
   * Not found.
   */
   return NULL;
}
