/******************************************************************************
 * $Id: treefunc.c,v 1.61 2001/04/14 03:06:58 jimb Exp $
 * Author		: Jim Brooks	<jim@jimbrooks.org>
 * Description	: Functions to support a tree view of a function call graph.
 ******************************************************************************/

#include "header.h"
#include "parser.h"

static GSList*		pGslistFuncCall		= NULL;		/* list of calls in one func def */
static char*		pModuleName			= NULL;		/* name of module being parsed */
static GHashTable*  pHashSubtrees		= NULL;

#if 0
#define DEBUG_ADD_TO_GRAPH_PRINTF
#endif

#define CHANGE_STYLE_TREE(pWidget) \
{ \
   ChangeStyleColorTextSel( pWidget, \
                            &colorClistWidgetTextFg,  \
                            &colorClistWidgetTextBg,  \
                            &colorClistWidgetSelFg,   \
                            &colorClistWidgetSelBg ); \
}

/******************************************************************************
 * Build a graph of functions.
 * (see types.h for an illustration of the graph)
 ******************************************************************************/
void
BuildFuncGraph( void )
{
   GList*	pLink		= NULL;
   char*	pBuf		= NULL;
   long		moduleSize  = 0;

#ifdef HYPERSRC_PEDANTIC
  /*
   * Prerequisites.
   */
   g_assert( pGlistTags );
   g_assert( pHashModules );
   g_assert( pGlistModuleNames );
#endif

  /*
   * Pass the contents of each module to the function call parser.
   */
   for ( pLink = pGlistModuleNames;
         pLink;
         pLink = pLink->next )
   {
      pBuf			= NULL;
      pModuleName	= (char*) pLink->data;
g_assert( pModuleName );

     /*
      * The parser only supports C/C++ source files.
      */
      if ( ! IsCSrcFile( pModuleName ) )
         continue;

#if defined(HYPERSRC_DEBUG) && defined(DEBUG_ADD_TO_GRAPH_PRINTF)
      DebugMsg( "\n*******************************************************************" );
      DebugMsg( "Parsing module %s ", pModuleName );
      DebugMsg( "*******************************************************************\n" );
#endif

     /*
      * Reset state before parsing another module.
      */
      pGslistFuncCall = NULL;

     /*
      * Read the contents of this module.
      */
      pBuf = ModuleBuffer( pModuleName, &moduleSize );
      if ( pBuf == NULL  ||  moduleSize <= 0 )
      {
         /* --module.c has already called Warning()--*/
         /*Warning( "Couldn't open module %s for parsing.", pModuleName );*/
         continue;	/* can happen if a src file is empty, eg Mesa3D has a few */
      }

     /*
      * Pass the contents of this module to the parser.
      */
      ParseFuncsInModule( pBuf, moduleSize, pModuleName, FALSE );
   }

#if 0
#if defined(HYPERSRC_DEBUG)  &&  !defined(DEBUG_ADD_TO_GRAPH_PRINTF)
   DumpFuncGraph();
#endif
#endif

   return;
}

/*****************************************************************************
 * This is called as soon as the identifier of a function call is parsed.
 * Appends a funcCall_t struct to a gslist (see types.h).
 * Later, when the function definition is completely parsed,
 * AddFuncDefToGraph() will be called which will attach the func call gslist
 * to a funcDef_t struct.
 *****************************************************************************/
void
AddFuncCallToGraph( char* pFuncCallId,
                    int   line )
{
   funcCall_t*	pFuncCall;
   PTAG			pTag;

#ifdef HYPERSRC_PEDANTIC
   g_assert( pModuleName  &&  pModuleName[0] );
   g_assert( pFuncCallId  &&  StrLen(pFuncCallId) < MAX_TAG_NAME_LEN );
   g_assert( line > 0  &&  line < MAX_LINES );
#endif

#if defined(HYPERSRC_DEBUG)  &&  defined(DEBUG_ADD_TO_GRAPH_PRINTF)
   /*---DEBUG---*/
   DebugMsg( "    calls %s() line # %d ", pFuncCallId, line );
   return;
   /*---DEBUG---*/
#endif

  /*
   * Try to find tag of function call (which was parsed).
   */
   pTag = FindFuncTag( pFuncCallId, pModuleName );
   if ( pTag == NULL )
   {
     /*
      * Is the function call really a function macro?
      * FindFuncTag() won't match a function macro.
      */
      pTag = FindTag( pFuncCallId, "macro" );
      if ( pTag == NULL )
      {
        /*
         * Tag wasn't found, so fabricate one.
         * In order to reduce the amount of tags appearing in the clist widget,
         * fabricated tags are not linked with other tags.
         */
         pTag = AllocTagStruct();
         FillTagStruct( pTag,
                        pFuncCallId,
                        line,
                        "call",
                        pModuleName );
      }
   }

  /*
   * Allocate a funcCall_t struct.
   */
   pFuncCall = (funcCall_t*) g_malloc( sizeof(funcCall_t) );

  /*
   * Fill struct.
   */
   pFuncCall->pTagFuncCall = pTag;

  /*
   * Append struct to func call gslist.
   * (This pointer must be NULLed before starting to parse a func def.)
   */
   pGslistFuncCall = g_slist_append( pGslistFuncCall, pFuncCall );

   return;
}

/*****************************************************************************
 * Appends a funcDef_t struct to a list for the module being parsed
 * (whose head is pointed to by this module's hash item).
 *
 * This is called when the parsing of a function definition completes,
 * which typically is after multiple calls to AddFuncCallToGraph().
 *
 * Parms	: pHashModules, pHashItemModule (OUT/GLOBAL)
 *            Updates hash item's "pGslistFuncDef" member.
 *            pHashItemModule points to hash item of module being parsed.
 *
 *            pGslistFuncCall (OUT/GLOBAL)
 *            NULLified before returning to prepare for next time.
 *****************************************************************************/
void
AddFuncDefToGraph( char* pFuncDefId,
                   int   lineLeftBrace,
                   int   lineRightBrace )
{
   funcDef_t*	pFuncDef;
   PTAG			pTag;

#ifdef HYPERSRC_PEDANTIC
   g_assert( pHashModules  &&  pHashItemModule  &&  pModuleName  &&  pModuleName[0] );
   g_assert( pFuncDefId  &&  StrLen(pFuncDefId) < MAX_TAG_NAME_LEN );
   g_assert( lineLeftBrace  > 0  &&  lineLeftBrace  < MAX_LINES );
   g_assert( lineRightBrace > 0  &&  lineRightBrace < MAX_LINES );
   g_assert( lineLeftBrace <= lineRightBrace );
#endif

#if defined(HYPERSRC_DEBUG)  &&  defined(DEBUG_ADD_TO_GRAPH_PRINTF)
   /*---DEBUG---*/
   DebugMsg( "  %s{} body resides at lines %d..%d \n", pFuncDefId, lineLeftBrace, lineRightBrace );
   /*---DEBUG---*/
#endif

  /*
   * Find tag of function (which was parsed as a function definition).
   */
   pTag = FindFuncTagInOneModule( pFuncDefId, pHashItemModule );
   if ( pTag )
   {
     /*
      * Allocate a funcDef_t struct.
      */
      pFuncDef = (funcDef_t*) g_malloc( sizeof(funcDef_t) );

     /*
      * Fill struct.
      */
      pFuncDef->pTagFuncDef		= pTag;
      pFuncDef->pGslistFuncCall = pGslistFuncCall;	/* might be NULL no other funcs are called */
      pFuncDef->lineStart		= lineLeftBrace;
      pFuncDef->lineLast		= lineRightBrace;

     /*
      * Append struct to the func def gslist whose pointer is stored
      * in an item in the module hash table.
      */
      pHashItemModule->pGslistFuncDef = \
      g_slist_append( pHashItemModule->pGslistFuncDef,
                      pFuncDef );
   }

  /*
   * Prepare for the next function definition.
   */
   pGslistFuncCall = NULL; /* so that g_slist_append() will create a new gslist */
   return;
}

/*****************************************************************************
 * Refresh the functree widget.
 * This is called when the verbosity pref is changed.
 *****************************************************************************/
void
RefreshFuncTree( void )
{
   int	line;

   if ( IsTextWidgetEmpty() ) return;
   line = PrintLineNum();
   UpdateFuncTree( line, TRUE );

   return;
}

/*****************************************************************************
 * Determine whether to update the functree per current cursor position.
 * This is invoked by an idle function, that is in turn invoked by
 * keypress event handlers.
 *
 * Parms	: line
 *            Line # of text cursor.
 *
 *            force
 *            Force update if possible.
 *
 * Notes    : The global pHashItemModule can't be used here because
 *            a global string search could make it reference another
 *            module that the text widget isn't showing.
 *****************************************************************************/
void
UpdateFuncTree( int      line,
                gboolean force )
{
   static P_HASH_ITEM_MODULE pHashItemModulePrev	= NULL;
   static int			linePrev			= -1;
   static funcDef_t*	pFuncDefPrev		= NULL;
   static funcDef_t*	pFuncDef			= NULL;
   static gboolean      busy                = FALSE;

  /*
   * Quit if check button is set.
   */
   if ( disableFuncTreeUpdates ) return;

  /*
   * Avoid nesting.
   */
   if ( busy  ||  skipFuncTree ) return;
   busy = TRUE;

  /*
   * Do nothing if the module and/or text widget appears empty.
   * Or line is invalid (-1).
   */
   if (    line < 0
        || ActiveModuleHashItem() == NULL
#if 0
        || ActiveModuleHashItem()->pModuleBuffer == NULL /* ignore, might change to NULL by global searches */
#endif
        || IsTextWidgetEmpty() )
      goto idle_quit;

  /*
   * Do nothing if the cursor is at the same spot (same module/line).
   */
   if (     !force
        &&  line == linePrev
        &&  ActiveModuleHashItem() == pHashItemModulePrev )
      goto idle_quit;

   linePrev = line;
   pHashItemModulePrev = ActiveModuleHashItem();

  /*
   * Search the list of funcDef_t structs to determine if the line
   * the cursor is on is within a function definition.
   */
   pFuncDef = SearchFuncDefList( line );

  /*
   * Is the cursor at the same function definition?
   */
   if ( !force  &&  pFuncDef == pFuncDefPrev )
     goto idle_quit;

   pFuncDefPrev = pFuncDef;

  /*
   * Actually update the tree widget.
   */
   UpdateFuncTreeWidget( pFuncDef );

idle_quit:
   busy = FALSE;
   return;
}

/*****************************************************************************
 * Create/update the tree widget that depicts function call relationships.
 *
 * Parms	: pFuncDef
 *            Pointer to a list of funcDef_t structs.
 *            If this is NULL then the tree widget will be cleared.
 *
 *            pWidgetFuncTree (OUT/GLOBAL)
 *
 * Notes    : If a corresponding func def/call list is empty,
 *            then the root of a subtree will still be created and shown
 *            (this makes sense if you hold Down thru a text widget while
 *             looking at the tree widget, you don't want it to skip a func def).
 *            If a subtree isn't shown, then that means the parser missed items.
 *****************************************************************************/
void
UpdateFuncTreeWidget( funcDef_t* pFuncDef )
{
   static GtkWidget*	pWidgetItemCalls		= NULL;
   static GtkWidget*	pWidgetItemSeparator	= NULL;
   static GtkWidget*	pWidgetItemCalledBy		= NULL;
		  GtkWidget*	pWidgetSubitem			= NULL;
          GtkWidget*	pWidgetSubtreeCalls		= NULL;
          GtkWidget*	pWidgetSubtreeCalledBy	= NULL;
          GSList*		pFuncCallList			= NULL;
          GSList*		pFuncDefList			= NULL;
          GList*		pLinkModuleName;
          char*			pFuncDefName			= NULL;
   static char			str[MAX_TAG_NAME_LEN];

   if ( skipFuncTree ) return;

  /*
   * Create or clear tree widget.
   */
   if ( pWidgetFuncTree == NULL )
   {
      pWidgetFuncTree = gtk_tree_new();
      CHANGE_STYLE_TREE( pWidgetFuncTree );
      gtk_widget_show( pWidgetFuncTree );
   }
   else
   {
     /*
      * First remove subtrees, then items on the first-level.
      */
      if ( pWidgetItemCalls )
      {
         gtk_tree_item_remove_subtree( GTK_TREE_ITEM(pWidgetItemCalls) );
         pWidgetItemCalls = NULL;
      }
      if ( pWidgetItemCalledBy )
      {
         gtk_tree_item_remove_subtree( GTK_TREE_ITEM(pWidgetItemCalledBy) );
         pWidgetItemCalledBy = NULL;
      }
      gtk_tree_remove_items( GTK_TREE(pWidgetFuncTree),
                             GTK_TREE(pWidgetFuncTree)->children );
   }

  /*
   * Return if there are no func defs.
   */
   if ( pFuncDef == NULL ) goto quit;

  /*
   * Create the "foo() calls:" item on the first-level.
   */
g_assert( pFuncDef->pTagFuncDef );
   pFuncDefName = pFuncDef->pTagFuncDef->pName;
   sprintf( str, "%s{} calls:",  pFuncDefName );

   pWidgetItemCalls = CreateTreeItem( pWidgetFuncTree, str, NULL );

  /*
   * Create a subtree to contain list of function calls.
   * (A signal connection will undo any previous signal connection,
   *  so gtk_signal_disconnect() is unnecessary.)
   */
   pWidgetSubtreeCalls = CreateSubtree( pWidgetItemCalls, TRUE, pFuncDef->pTagFuncDef, NULL  );

   for ( pFuncCallList = pFuncDef->pGslistFuncCall;
         pFuncCallList;
         pFuncCallList = pFuncCallList->next )
   {
      funcCall_t*	pFuncCall;
      char*			pFuncCallName;

      pFuncCall     = pFuncCallList->data;		/* pointer to funcCall per se */
      pFuncCallName = pFuncCall->pTagFuncCall->pName;
      if ( IgnoreFuncName( pFuncCallName ) ) continue;

     /*
      * Grow a leaf with the name of this func def.
      */
      sprintf( str, "%s()", pFuncCallName );
      pWidgetSubitem = CreateTreeItem( pWidgetSubtreeCalls, str, pFuncCall->pTagFuncCall );
   }
   gtk_tree_item_expand( GTK_TREE_ITEM(pWidgetItemCalls) );

  /*
   * Create a separator item on the first-level, for readability.
   */
   pWidgetItemSeparator = CreateTreeItem( pWidgetFuncTree, "--------------------------", NULL );

  /*
   * Create the "foo() called by:" item on the first-level.
   */
   sprintf( str, "%s{} is called by:",  pFuncDefName );
   pWidgetItemCalledBy = CreateTreeItem( pWidgetFuncTree, str, NULL );

  /*
   * Create the "called by" subtree.
   */
   pWidgetSubtreeCalledBy = CreateSubtree( pWidgetItemCalledBy, TRUE, pFuncDef->pTagFuncDef, NULL );

   for ( pLinkModuleName = pGlistModuleNames;
         pLinkModuleName;
         pLinkModuleName = pLinkModuleName->next )
   {
      P_HASH_ITEM_MODULE pHashItemModuleTmp;

      pHashItemModuleTmp = (P_HASH_ITEM_MODULE) \
         g_hash_table_lookup( pHashModules, (gconstpointer)pLinkModuleName->data );

g_assert( pHashItemModuleTmp );

      for ( pFuncDefList = pHashItemModuleTmp->pGslistFuncDef;
            pFuncDefList;
            pFuncDefList = pFuncDefList->next )
      {
         for ( pFuncCallList = ((funcDef_t*)pFuncDefList->data)->pGslistFuncCall;
               pFuncCallList;
               pFuncCallList = pFuncCallList->next )
         {
            funcCall_t* pFuncCall = pFuncCallList->data; /* pointer to funcCall per se */
            char*		pCallerName;

g_assert( pFuncCall );
g_assert( pFuncDefList->data );

           /*
            * Is this function call calling the function definition?
            */
            if ( CompareStrings( pFuncDefName, pFuncCall->pTagFuncCall->pName ) )
               continue; /* doesn't match */

           /*
            * To reduce verbosity, avoid displaying multiple calls by the SAME tagged function.
            * However, do show each call from same-named functions in different modules.
            */
            if ( FilterDuplicateFuncCalls( ((funcDef_t*)pFuncDefList->data)->pGslistFuncCall, pFuncCall ) )
               continue;

           /*
            * Grow a leaf with the caller's name.
            */
            pCallerName = ((funcDef_t*)pFuncDefList->data)->pTagFuncDef->pName;
g_assert( pCallerName );
            if ( !funcTreeVerbose )
               sprintf( str, "%s()", pCallerName );
            else
               sprintf( str, "%s()  [%s]", pCallerName, (char*)pLinkModuleName->data );

            pWidgetSubitem = CreateTreeItem( pWidgetSubtreeCalledBy,
                                             str,
                                             ((funcDef_t*)pFuncDefList->data)->pTagFuncDef );
         }         
      }
   }
   gtk_tree_item_expand( GTK_TREE_ITEM(pWidgetItemCalledBy) );

  /*
   * Unless already done, create the functree's ancillary widgets.
   */
quit:
   if ( pWidgetFrameFuncTree )
      return;

   CreateFuncTreeAncillaryWidgets();
   return;
}

/*****************************************************************************
 * Create the functree's ancillary widgets.
 * Later, the functree's frame will be added to a paned widget (along with the notebook).
 *
 * Parms	: pWidgetFrameFuncTree (OUT/GLOBAL)
 *
 *            pWidgetScrolledFuncTree (OUT/GLOBAL)
 *
 *            pWidgetCheckButtonFunctree (OUT/GLOBAL)
 *****************************************************************************/
void
CreateFuncTreeAncillaryWidgets( void )
{
   GtkWidget*	pWidgetLabel	= NULL;
   GtkWidget*	pWidgetHbox		= NULL;
   GtkWidget*	pWidgetVbox		= NULL;

#undef  SPACING 
#define SPACING		1
#undef  PADDING
#define PADDING		1

  /*
   * Create outer-level vbox.
   */
   pWidgetVbox = gtk_vbox_new( FALSE, SPACING );

  /*
   * Create a scrolled widget for the functree.
   */
   pWidgetScrolledFuncTree = gtk_scrolled_window_new( NULL, NULL );
   gtk_scrolled_window_add_with_viewport( GTK_SCROLLED_WINDOW(pWidgetScrolledFuncTree),
                                          pWidgetFuncTree );

  /*
   * Pack scrolled widget into vbox.
   */
   gtk_box_pack_start( GTK_BOX(pWidgetVbox),
                       pWidgetScrolledFuncTree,
                       TRUE, TRUE, PADDING );
  /*
   * Create check button to disable functree, pack it into vbox.
   */
   pWidgetHbox = gtk_hbox_new( FALSE, SPACING );				/* create hbox */

   pWidgetLabel = gtk_label_new( "             " );
   gtk_box_pack_start( GTK_BOX(pWidgetHbox),					/* pack label in hbox */
                       pWidgetLabel,
                       FALSE, FALSE, PADDING );

   pWidgetCheckButtonFunctree = gtk_check_button_new();			/* create check button */
   gtk_box_pack_start( GTK_BOX(pWidgetHbox),					/* pack check button in hbox */
                       pWidgetCheckButtonFunctree,
                       FALSE, FALSE, PADDING );

   pWidgetLabel = gtk_label_new( "stop updating tree" );
   gtk_box_pack_start( GTK_BOX(pWidgetHbox),					/* pack label in hbox */
                       pWidgetLabel,
                       FALSE, FALSE, PADDING );

   gtk_box_pack_start( GTK_BOX(pWidgetVbox),					/* pack hbox in vbox */
                       pWidgetHbox,
                       FALSE, FALSE, PADDING );

   gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(pWidgetCheckButtonFunctree),
                                 FALSE );						/* before connecting signal! */
   gtk_signal_connect( GTK_OBJECT(pWidgetCheckButtonFunctree),	/* connect toggle signal */
                       "toggled",
                       GTK_SIGNAL_FUNC(HandlerButtons),
                       NULL );

  /*
   * Finally, add the vbox into the functree's frame.
   */
   pWidgetFrameFuncTree = gtk_frame_new( NULL );
   gtk_container_add( GTK_CONTAINER(pWidgetFrameFuncTree),		/* pack vbox in frame */
                      pWidgetVbox );

  /*
   * Install focus handlers.
   */
   InstallFocusHandlers( pWidgetScrolledFuncTree );

   return;
}

/*****************************************************************************
 * Returns true if a function name should not be added to a functree.
 *****************************************************************************/
gboolean
IgnoreFuncName( char* pFuncName )
{
  /*
   * Ignore C keywords mistaken as function names.
   */
   if (    CompareStrings( pFuncName, "if"      ) == 0
        || CompareStrings( pFuncName, "for"     ) == 0
        || CompareStrings( pFuncName, "while"   ) == 0
        || CompareStrings( pFuncName, "switch"  ) == 0
        || CompareStrings( pFuncName, "return"  ) == 0
        || CompareStrings( pFuncName, "sizeof"  ) == 0
        || CompareStrings( pFuncName, "defined" ) == 0 )
      return TRUE;

   return FALSE;
}

/*****************************************************************************
 * Filters out later duplicate calls made by the same func def.
 *
 * Parms	: pFuncCallList
 *            List of funcCall_t structs.  This list belongs to one func def.
 *
 *            pFuncCall
 *            Must point to a link in pFuncCallList.
 *
 * Returns	: Returns true if this func call should be ignored.
 *            Returns false if no duplicate calls precedes the referenced one.
 *****************************************************************************/
gboolean
FilterDuplicateFuncCalls( GSList*     pFuncCallList,
                          funcCall_t* pFuncCall )
{
#ifdef HYPERSRC_PEDANTIC
g_assert( pFuncCallList  &&  pFuncCall );
#endif

  /*
   * Only return false for the first duplicated call (there will be at least one).
   */
   for ( ;
         pFuncCallList;
         pFuncCallList = pFuncCallList->next )
   {
      if ( ((funcCall_t*)pFuncCallList->data)->pTagFuncCall == pFuncCall->pTagFuncCall )
         return pFuncCallList->data != pFuncCall;
   }

  /*
   * Shouldn't reach here.
   */
   g_assert(0);
   return TRUE;
}

/*****************************************************************************
 * Search the list of function definitions to determine if the line
 * the cursor is on is within a function definition.
 *
 * Parms	: line
 *            Line # of text cursor.
 *
 * Returns	: Pointer to a funcDef_t struct.
 *****************************************************************************/
funcDef_t*
SearchFuncDefList( int line )
{
   GSList*		pGslistFuncDef;
   int			lineStart;
   int			lineLast;

#ifdef HYPERSRC_PEDANTIC
g_assert( ActiveModuleHashItem() );
#endif

  /*
   * For each funcDef struct in the list.
   */
   for ( pGslistFuncDef = ActiveModuleHashItem()->pGslistFuncDef;
         pGslistFuncDef;
         pGslistFuncDef = pGslistFuncDef->next )
   {
      funcDef_t*	pFuncDef = pGslistFuncDef->data; /* pointer to funcDef per se */

      lineStart = pFuncDef->lineStart;
      lineLast  = pFuncDef->lineLast;

#ifdef HYPERSRC_PEDANTIC
      if ( ! ( lineStart <= lineLast  &&  lineStart >=0  &&  lineLast >= 0  ) )
      {
         DebugMsg( "SearchFuncDefList() line checks failed in module %s func def %s", pModuleName, pFuncDef->pTagFuncDef->pName );
         DebugMsg( "  lineStart = %d  lineLast %d ", lineStart, lineLast );
         continue;
      }
#endif

      if ( line >= lineStart  &&  line <= lineLast )
         return pFuncDef;
   }

   return NULL;
}

/*****************************************************************************
 * Create notebook page that shows all functions in all modules.
 * The modules will be sorted by name, while functions are sorted by line.
 *****************************************************************************/
void
CreateNotebookFuncTree( void )
{
   GtkWidget*	pWidgetLabel;
   GList*		pLinkModuleName;

   if ( skipFuncTree ) return;

#ifdef HYPERSRC_PEDANTIC
   g_assert( pWidgetNotebookFuncTree == NULL  &&  pHashModules  &&  pWidgetNotebook );
#endif

  /*
   * Create hash to remember that module pathname associated with a subtree.
   */
   pHashSubtrees = g_hash_table_new( g_str_hash, g_str_equal );
g_assert( pHashSubtrees );

  /*
   * Create tree widget (for a notebook page).
   */
   pWidgetNotebookFuncTree = gtk_tree_new();
   CHANGE_STYLE_TREE( pWidgetNotebookFuncTree );

  /*
   * For each module (using sorted list of module names).
   */
   for ( pLinkModuleName = pGlistModuleNames;
         pLinkModuleName;
         pLinkModuleName = pLinkModuleName->next )
   {
      P_HASH_ITEM_MODULE pHashItemModuleTmp;

      pHashItemModuleTmp = (P_HASH_ITEM_MODULE) \
         g_hash_table_lookup( pHashModules, (gconstpointer)pLinkModuleName->data );
g_assert( pHashItemModuleTmp );
      NotebookFuncTreeForEachModule( pLinkModuleName->data,
                                     pHashItemModuleTmp );
   }

  /*
   * Add this tree widget to the notebook.
   */
   pWidgetNotebookScrolledFuncTree = gtk_scrolled_window_new( NULL, NULL );
   gtk_scrolled_window_add_with_viewport( GTK_SCROLLED_WINDOW(pWidgetNotebookScrolledFuncTree),
                                          pWidgetNotebookFuncTree );

   pWidgetLabel = gtk_label_new( "functions" );
   gtk_notebook_append_page( GTK_NOTEBOOK(pWidgetNotebook),
                             pWidgetNotebookScrolledFuncTree,
                             pWidgetLabel );

   gtk_widget_show( pWidgetNotebookFuncTree );
   gtk_widget_show( pWidgetNotebookScrolledFuncTree );

  /*
   * Free subtree hash.
   */
   g_hash_table_foreach( pHashSubtrees, FreeHashSubtreeItem, NULL ); /* free each dir string */
   g_hash_table_destroy( pHashSubtrees );
   pHashSubtrees = NULL;

   return;
}

/*****************************************************************************
 * For a module, grows a tree branch corresponding to the module's pathname,
 * and grows leafs for each function in the module.
 *****************************************************************************/
void
NotebookFuncTreeForEachModule( char*              pModulePathname,
                               P_HASH_ITEM_MODULE pModuleHashItem )
{
   GSList*				pFuncDefList;
   char*				pFuncDefName;
   GtkWidget*			pWidgetItem					= NULL;
   GtkWidget*			pWidgetParentTreeFilename	= NULL;		/* parent tree of subtree labeled with module's filename */
   GtkWidget*			pWidgetSubtree				= NULL;
   GtkWidget*			pWidgetSubitem				= NULL;
   static char			str[MAX_TAG_NAME_LEN];

#ifdef HYPERSRC_PEDANTIC
   g_assert( pModulePathname && pModuleHashItem );
#endif

  /*
   * Create a tree item labeled with this module's filename.
   * But first, a node must be created for each directory component
   * in the module's pathname.
   */
   pWidgetParentTreeFilename = DescendTreeNodesOnModulePath( pModulePathname );
   pWidgetItem = CreateTreeItem( pWidgetParentTreeFilename,
                                 ExtractFilenameFromPathname( pModulePathname ),
                                 NULL );
  /*
   * Create a subtree (to hold leafs with names of functions in this module).
   */
   pWidgetSubtree = CreateSubtree( pWidgetItem, TRUE, NULL, pModulePathname );

  /*
   * For each tagged function in this module.
   */
   for ( pFuncDefList = pModuleHashItem->pGslistFuncDef;
         pFuncDefList;
         pFuncDefList = pFuncDefList->next )
   {
     /*
      * Grow a leaf with this function's name.
      */
      pFuncDefName = ((funcDef_t*)(pFuncDefList->data))->pTagFuncDef->pName;
      sprintf( str, "%s()", pFuncDefName );
      pWidgetSubitem = CreateTreeItem( pWidgetSubtree,
                                       str,
                                       ((funcDef_t*)pFuncDefList->data)->pTagFuncDef );
   }

   return;
}

/*****************************************************************************
 * Prior to destroying the subtree hash, free memory used by each hash item.
 *****************************************************************************/
void
FreeHashSubtreeItem( gpointer key,
                     gpointer value,
                     gpointer unused )
{
  /*
   * Free the hash key's string which was created by CloneString().
   */
   DeleteString( key );
}

/*****************************************************************************
 * For the notebook functree, descend thru the tree nodes that correspond
 * to each directory on the path to a module, creating any node if necessary.
 *
 * The caller is expected to create a subtree for the module's filename,
 * and populate it with function names.
 *
 * Parms    : pModulePathname
 *
 *            pWidgetNotebookFuncTree (GLOBAL)
 *            Affected, must be created prior.
 *
 *            pHashSubtrees (GLOBAL)
 *            Affected, must be created prior.
 *
 * Returns	: A pointer to the tree that can be the parent of the subtree
 *            that would contain the module's filename.
 *            The returned tree is either the root or of the last dir part.
 *****************************************************************************/
GtkWidget*
DescendTreeNodesOnModulePath( char* pModulePathname )
{
   char*    pNextDir;				/* tentatively points to next dir part */
   char*	pDirBegin;
   char*	pDirEnd;
   int      len;
   char     key[MAX_STRING_LEN];
   char     dir[MAX_STRING_LEN];
   gpointer pHashItem;
   int      iteration;
   GtkWidget* pWidgetParentTree;
   GtkWidget* pWidgetSubtree;
   GtkWidget* pWidgetItem;

#ifdef HYPERSRC_PEDANTIC
   g_assert( pModulePathname  &&  pWidgetNotebookFuncTree  &&  pHashSubtrees );
#endif

   pWidgetParentTree = pWidgetNotebookFuncTree;

  /*
   * For each dir in the module's pathname.
   */
   pNextDir = pModulePathname;
   for ( iteration = 0;
         ;
         ++iteration )
   {
     /*
      * Try to extract the next directory part from the module's pathname.
      */
      ExtractDirFromPathname( pNextDir,
                              &pDirBegin,
                              &pDirEnd );

     /*
      * Break if no dir part was extracted (ie remainder is a filename).
      */
      if ( pDirBegin == NULL ) break;

     /*
      * Tentatively update the pointer to the next dir part.
      */
      pNextDir = pDirEnd + 1;

     /*
      * Has a subtree been created that corresponds
      * to the path that has so far been extracted?
      * If so then continue, else create a subtree.
      */
      len = pDirEnd - pModulePathname + 1;
#ifdef HYPERSRC_PEDANTIC
      g_assert( len > 0 );
#endif
      memcpy( key, pModulePathname, len );
      key[len] = '\0';
      pHashItem = g_hash_table_lookup( pHashSubtrees, key );

      if ( pHashItem )
      {
        /*
         * Update pointer to parent tree for next iteration.
         */
         pWidgetParentTree = (GtkWidget*) pHashItem; /* hash data is pointer to subtree */
         continue;
      }

     /*
      * Create a subtree whose label is the extracted dir part of the module's pathname.
      */
      len = pDirEnd - pDirBegin + 1;		/* stringify the dir part */
#ifdef HYPERSRC_PEDANTIC
      g_assert( len > 0 );
#endif
      memcpy( dir, pDirBegin, len );
      dir[len]   = '/';						/* append '/' to show it is a dir, not a src file */
      dir[len+1] = '\0';

      pWidgetItem = CreateTreeItem( pWidgetParentTree, dir, NULL );
      pWidgetSubtree = CreateSubtree( pWidgetItem, FALSE, NULL, NULL );

     /*
      * Remember that this subtree was created by inserting
      * the extracted path into a hash.
      */
      g_hash_table_insert( pHashSubtrees,
#if 0
                           (gpointer) key,	/* <-- ### WRONG: key's string ephemerally resides on local stack. */
#else                                       /*     ### Must explicitly allocate key's string. */
                           (gpointer) CloneString(key),
#endif
                           (gpointer) pWidgetSubtree );

     /*
      * Update the parent tree pointer, and reiterate.
      */
      pWidgetParentTree = pWidgetSubtree;
   }

  /*
   * Return the last update to the parent tree's pointer.
   * If the pathname had no dirs then this will be the true root tree.
   */
   return pWidgetParentTree;
}

/*****************************************************************************
 * Primitive to create a subtree (node).
 *
 * Parms	: pWidgetItem
 *            Pointer to tree item to be transformed as a node.
 *
 *            connectSignals
 *            Whether to connect to signals when the subtree (node) or
 *            its items (leafs) are selected.
 *
 *            pTag or pModule (either, not both)
 *            Either a tag or module name can be attached as object data
 *            to the tree widget for use by a handler.
 *            Only one pointer can be passed, the other must be passed as NULL.
 *            (Note: the object data must be attached to its tree item form).
 *****************************************************************************/
GtkWidget*
CreateSubtree( GtkWidget* pWidgetItem,
               gboolean   connectSignals,
               PTAG       pTag,
               char*      pModule )
{
   GtkWidget* pWidgetSubtree;

#ifdef HYPERSRC_PEDANTIC
g_assert( pWidgetItem );
g_assert( ! (pTag && pModule) );	/* either, not both */
#endif

  /*
   * Create a subtree.
   */
   pWidgetSubtree = gtk_tree_new();
#ifdef HYPERSRC_PEDANTIC
g_assert( pWidgetSubtree );
#endif
   CHANGE_STYLE_TREE( pWidgetSubtree );
   gtk_tree_item_set_subtree( GTK_TREE_ITEM(pWidgetItem), pWidgetSubtree );

  /*
   * If requested, connect to signals.
   */
   if ( connectSignals )
   {
     /*
      * Attach object data for use by handler.
      * (Note: attaching data to pWidgetSubtree didn't work).
      */
      if ( pTag )
         gtk_object_set_data( GTK_OBJECT(pWidgetItem), "ptag",   pTag );
      else if ( pModule )
         gtk_object_set_data( GTK_OBJECT(pWidgetItem), "module", pModule );

     /*
      * For when an item of the subtree is selected.
      * Connect to the GtkTree "select_child" signal.
      */
      gtk_signal_connect( GTK_OBJECT(pWidgetSubtree),	/* yes, pWidgetSubtree */
                          "select_child",
                          (GtkSignalFunc) HandlerTreeItemSelected,
                          NULL );

     /*
      * For when the subtree itself is selected (ie the node that can be expanded/collapsed).
      * Connect to the GtkItem "select" signal.
      */
      gtk_signal_connect( GTK_OBJECT(pWidgetItem),		/* yes, pWidgetItem */
                          "select",
                          (GtkSignalFunc) HandlerSubtreeSelected,
                          NULL );
   }

   return pWidgetSubtree;
}

/*****************************************************************************
 * Primitive to create a tree item (leaf).
 * Optionally, the leaf's object data can be assigned a pointer to a TAG struct.
 *****************************************************************************/
GtkWidget*
CreateTreeItem( GtkWidget* pWidgetParentTree,
                char*      pItemName,
                PTAG       pTag )
{
   GtkWidget* pWidgetItem;

   pWidgetItem = gtk_tree_item_new_with_label( pItemName );
   CHANGE_STYLE_TREE( pWidgetItem );
   gtk_tree_append( GTK_TREE(pWidgetParentTree), pWidgetItem );
   gtk_widget_show( pWidgetItem );

   if ( pTag ) gtk_object_set_data( GTK_OBJECT(pWidgetItem), "ptag", pTag );

   return pWidgetItem;
}

/*****************************************************************************
 * Show the line of src code corresponding to a selected tree node or item.
 *****************************************************************************/
gint
IdleCallback_HyperjumpViaTreeNodeOrItem( gpointer pTag )
{
   HyperjumpViaTreeNodeOrItem( (PTAG) pTag );
   return FALSE;  /* FALSE tells GTK+ to stop calling an idle callback */
}
void
HyperjumpViaTreeNodeOrItem( PTAG pTag )
{
   static gboolean	busy					= FALSE;
   int				row						= -1;
   GtkWidget*		pWidgetTextActivePrev   = pWidgetTextActive;

   if ( pTag == NULL )
   {
      DebugMsg( __FUNCTION__"(): pTag==NULL" );
      return;
   }

  /*
   * Prevent nesting.
   */
   if ( busy  ||  IsAnyMajorFunctionBusy() ) return;
   busy = TRUE;

   gtk_clist_freeze( GTK_CLIST( pWidgetClistTags ) );
   gtk_clist_freeze( GTK_CLIST( pWidgetClistModules ) );

  /*
   * Record the current cursor position into the tags history.
   * This tag is needed so that the user can jump back to previous cursor positions
   * (via Prev Tag button).
   */
   RecordCursorInTagHistory( "(recorded position)" );

  /*
   * Should the tagged line be shown in the main or aux text widget?
   */
   if ( ! auxKeyPressed )
      SetActiveTextWidget( pWidgetTextMain );
   else
      SetActiveTextWidget( pWidgetTextAux );

   row = \
   FindTagCheckMultipleMatches( pTag->pName,
                                pTag->pModule,
                                widgetTagsClistSelectedRow,
                                TRUE,	/* be exact */
                                TRUE,   /* (case N/A) */
                                FALSE,  /* don't flip to another notebook */
                                "Multiple tags share the same name (press F5 for next one)." );

   if (    (tagsCnt > TAGS_VERY_MANY)
        && GetStatusbarMsg()
        && CompareStrings( GetStatusbarMsg(), "Going to tag..." ) == 0 )
      MostlyClearStatusbar();

  /*
   * If the search failed then remove the tag of the current position
   * because since no hyperlink was done, the current position won't change.
   */
   if ( row == ROW_NONE)
   {
      UndoRecordCursorInTagHistory();
      PrintStatusbarThirds( "'", pTag->pName, "' is not tagged" );
   }

   gtk_clist_thaw( GTK_CLIST( pWidgetClistTags ) );
   gtk_clist_thaw( GTK_CLIST( pWidgetClistModules ) );

  /*
   * Restore the active text widget.
   */
   pWidgetTextActive = pWidgetTextActivePrev;
   SetActiveTextWidget( pWidgetTextActive );

   busy = FALSE;
   return;
}

/*****************************************************************************
 * Load text widget with a module corresponding to the subtree (node)
 * selected in the notebook's functree.
 *****************************************************************************/
gint
IdleCallback_HyperjumpToModuleViaNotebookSubtree( gpointer pModule )
{
   HyperjumpToModuleViaNotebookSubtree( (char*) pModule );
   return FALSE;  /* FALSE tells GTK+ to stop calling an idle callback */
}
void
HyperjumpToModuleViaNotebookSubtree( char* pModule )
{
   static gboolean	busy					= FALSE;
   GtkWidget*		pWidgetTextActivePrev   = pWidgetTextActive;

#ifdef HYPERSRC_PEDANTIC
g_assert( pModule );
#endif
  /*
   * --- code was copied/pasted from list.c ---
   */

  /*
   * Prevent nesting.
   */
   if ( busy  ||  IsAnyMajorFunctionBusy() )
      return;
   busy = TRUE;

  /*
   * Should the module be shown in the main or aux text widget?
   */
   SetActiveTextWidget( !auxKeyPressed ? pWidgetTextMain : pWidgetTextAux );

  /*
   * Load the selected module (ie source file) into the text widget.
   * Move the cursor to the first line.
   */
   if ( LoadTextWidgetWithModule( pModule,
                                  0,
                                  ROW_NONE,
                                  DONT_RELOAD )
        != FUNC_OK )
   {
     /*
      * Oops, unable to load the module.
      * (A warning has already been displayed.)
      */
      goto quit;
   }

  /*
   * Clear statusbar.
   */
   MostlyClearStatusbar();

quit:
  /*
   * Restore the active text widget.
   */
   SetActiveTextWidget( pWidgetTextActive = pWidgetTextActivePrev );

   busy = FALSE;
   return;
}

/*-------------------------------------------------------------------------------
 * DEBUG
 *-------------------------------------------------------------------------------*/


#ifdef HYPERSRC_DEBUG

/*****************************************************************************
 * For debugging, print func graph.
 *****************************************************************************/
void
DumpFuncGraph( void )
{
   g_hash_table_foreach( pHashModules, CallbackDumpFuncGraph, NULL );
}
void
CallbackDumpFuncGraph( gpointer key,
                       gpointer val,
                       gpointer unused )
{
   P_HASH_ITEM_MODULE	pHashItem = (P_HASH_ITEM_MODULE) val;
   GSList*				pGslistFuncDefDbg	= NULL;
   GSList*				pGslistFuncCallDbg	= NULL;

  /*
   * This callback is passed a hash item from the module hash table.
   */
   DebugMsg( "\n------- module '%s' ---------------------------------------",
             pHashItem->pModuleName );

  /*
   * Get start of list of funcDef structs (from hash item).
   */
   pGslistFuncDefDbg = pHashItem->pGslistFuncDef;
   if ( pGslistFuncDefDbg == NULL )
   {
      DebugMsg( "(no function definitions were parsed or tags weren't found)\n" );
      return;
   }

  /*
   * For each funcDef struct in the list.
   */
   for ( ; pGslistFuncDefDbg; pGslistFuncDefDbg = pGslistFuncDefDbg->next )
   {
      funcDef_t*	pFuncDef = pGslistFuncDefDbg->data; /* pointer to funcDef per se */

      DebugMsg( "\n  %s(){..} defined at lines %d..%d",
                pFuncDef->pTagFuncDef->pName,
                pFuncDef->lineStart,
                pFuncDef->lineLast );

     /*
      * Get start of list of funcCall structs (pointed to by this funcDef struct).
      */
      pGslistFuncCallDbg = pFuncDef->pGslistFuncCall;
      if ( pGslistFuncCallDbg == NULL )
      {
         DebugMsg( "      (no func calls were parsed or tags weren't found)" );
         continue;
      }

     /*
      * For each parsed func call.
      */
      for ( ; pGslistFuncCallDbg; pGslistFuncCallDbg = pGslistFuncCallDbg->next )
      {
         funcCall_t* pFuncCall = pGslistFuncCallDbg->data; /* pointer to funcCall per se */

         DebugMsg( "      %s() called",
                   pFuncCall->pTagFuncCall->pName );
      }
   }
}

#endif /* HYPERSRC_DEBUG */
